# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
from io import StringIO
from operator import itemgetter, attrgetter

from pyomo.common.config import (
    ConfigBlock,
    ConfigValue,
    InEnum,
    document_kwargs_from_configdict,
)
from pyomo.common.deprecation import deprecation_warning
from pyomo.common.gc_manager import PauseGC
from pyomo.common.timing import TicTocTimer

from pyomo.core.base import (
    Block,
    Objective,
    Constraint,
    Var,
    Param,
    Expression,
    SOSConstraint,
    Suffix,
    SymbolMap,
    minimize,
)
from pyomo.core.base.label import LPFileLabeler, NumericLabeler
from pyomo.opt import WriterFactory
from pyomo.repn.linear import LinearRepnVisitor
from pyomo.repn.quadratic import QuadraticRepnVisitor
from pyomo.repn.util import (
    FileDeterminism,
    FileDeterminism_to_SortComponents,
    OrderedVarRecorder,
    categorize_valid_components,
    initialize_var_map_from_column_order,
    int_float,
    ordered_active_constraints,
    row_order2row_map,
)

### FIXME: Remove the following as soon as non-active components no
### longer report active==True
from pyomo.core.base import Set, RangeSet, ExternalFunction
from pyomo.network import Port

logger = logging.getLogger(__name__)
inf = float('inf')
neg_inf = float('-inf')


# TODO: make a proper base class
class LPWriterInfo:
    """Return type for LPWriter.write()

    Attributes
    ----------
    symbol_map: SymbolMap

        The :py:class:`SymbolMap` bimap between row/column labels and
        Pyomo components.

    """

    def __init__(self, symbol_map):
        self.symbol_map = symbol_map


@WriterFactory.register(
    'cpxlp_v2', 'Generate the corresponding CPLEX LP file (version 2).'
)
@WriterFactory.register('lp_v2', 'Generate the corresponding LP file (version 2).')
class LPWriter:
    CONFIG = ConfigBlock('lpwriter')
    CONFIG.declare(
        'show_section_timing',
        ConfigValue(
            default=False,
            domain=bool,
            description='Print timing after writing each section of the LP file',
        ),
    )
    CONFIG.declare(
        'skip_trivial_constraints',
        ConfigValue(
            default=False,
            domain=bool,
            description='Skip writing constraints whose body is constant',
        ),
    )
    CONFIG.declare(
        'file_determinism',
        ConfigValue(
            default=FileDeterminism.ORDERED,
            domain=InEnum(FileDeterminism),
            description='How much effort to ensure file is deterministic',
            doc="""
            How much effort do we want to put into ensuring the
            LP file is written deterministically for a Pyomo model:

               - NONE (0) : None
               - ORDERED (10): rely on underlying component ordering (default)
               - SORT_INDICES (20) : sort keys of indexed components
               - SORT_SYMBOLS (30) : sort keys AND sort names (not declaration order)

            """,
        ),
    )
    CONFIG.declare(
        'symbolic_solver_labels',
        ConfigValue(
            default=False,
            domain=bool,
            description='Write variables/constraints using model names',
            doc="""
            Export variables and constraints to the LP file using human-readable
            text names derived from the corresponding Pyomo component names.
            """,
        ),
    )
    CONFIG.declare(
        'row_order',
        ConfigValue(
            default=None,
            description='Preferred constraint ordering',
            doc="""
            List of constraints in the order that they should appear in the
            LP file.  Unspecified constraints will appear at the end.""",
        ),
    )
    CONFIG.declare(
        'column_order',
        ConfigValue(
            default=None,
            description='Preferred variable ordering',
            doc="""
            List of variables in the order that they should appear in
            the LP file.  Note that this is only a suggestion, as the LP
            file format is row-major and the columns are inferred from
            the order in which variables appear in the objective
            followed by each constraint.""",
        ),
    )
    CONFIG.declare(
        'labeler',
        ConfigValue(
            default=None,
            description='Callable to use to generate symbol names in LP file',
            doc="""
            Export variables and constraints to the LP file using human-readable
            text names derived from the corresponding Pyomo component names.
            """,
        ),
    )
    CONFIG.declare(
        'output_fixed_variable_bounds',
        ConfigValue(
            default=False,
            domain=bool,
            description='DEPRECATED option from LPv1 that has no effect in the LPv2',
        ),
    )
    CONFIG.declare(
        'allow_quadratic_objective',
        ConfigValue(
            default=True,
            domain=bool,
            description='If True, allow quadratic terms in the model objective',
        ),
    )
    CONFIG.declare(
        'allow_quadratic_constraint',
        ConfigValue(
            default=True,
            domain=bool,
            description='If True, allow quadratic terms in the model constraints',
        ),
    )

    def __init__(self):
        self.config = self.CONFIG()

    def __call__(self, model, filename, solver_capability, io_options):
        if filename is None:
            filename = model.name + ".lp"

        # Duplicate io_options to avoid side-effects
        io_options = dict(io_options)
        # Map old solver capabilities to new writer options
        qp = solver_capability('quadratic_objective')
        if 'allow_quadratic_objective' not in io_options:
            io_options['allow_quadratic_objective'] = qp
        qc = solver_capability('quadratic_constraint')
        if 'allow_quadratic_constraint' not in io_options:
            io_options['allow_quadratic_constraint'] = qc

        with open(filename, 'w', newline='') as FILE:
            info = self.write(model, FILE, **io_options)
        return filename, info.symbol_map

    @document_kwargs_from_configdict(CONFIG)
    def write(self, model, ostream, **options):
        """Write a model in LP format.

        Returns
        -------
        LPWriterInfo

        Parameters
        ----------
        model: ConcreteModel
            The concrete Pyomo model to write out.

        ostream: io.TextIOBase
            The text output stream where the LP "file" will be written.
            Could be an opened file or a io.StringIO.

        """
        config = self.config(options)

        if config.output_fixed_variable_bounds:
            deprecation_warning(
                "The 'output_fixed_variable_bounds' option to the LP "
                "writer is deprecated and is ignored by the lp_v2 writer."
            )

        # Pause the GC, as the walker that generates the compiled LP
        # representation generates (and disposes of) a large number of
        # small objects.
        with PauseGC():
            return _LPWriter_impl(ostream, config).write(model)


class _LPWriter_impl:
    def __init__(self, ostream, config):
        self.ostream = ostream
        self.config = config
        self.symbol_map = None

    def write(self, model):
        timing_logger = logging.getLogger('pyomo.common.timing.writer')
        timer = TicTocTimer(logger=timing_logger)
        with_debug_timing = (
            timing_logger.isEnabledFor(logging.DEBUG) and timing_logger.hasHandlers()
        )

        ostream = self.ostream

        labeler = self.config.labeler
        if labeler is None:
            if self.config.symbolic_solver_labels:
                labeler = LPFileLabeler()
            else:
                labeler = NumericLabeler('x')
        self.symbol_map = SymbolMap(labeler)
        addSymbol = self.symbol_map.addSymbol
        aliasSymbol = self.symbol_map.alias
        getSymbol = self.symbol_map.getSymbol

        self.sorter = sorter = FileDeterminism_to_SortComponents(
            self.config.file_determinism
        )
        component_map, unknown = categorize_valid_components(
            model,
            active=True,
            sort=sorter,
            valid={
                Block,
                Constraint,
                Var,
                Param,
                Expression,
                # FIXME: Non-active components should not report as Active
                ExternalFunction,
                Set,
                RangeSet,
                Port,
                # TODO: Piecewise, Complementarity
            },
            targets={Suffix, SOSConstraint, Objective},
        )
        if unknown:
            raise ValueError(
                "The model ('%s') contains the following active components "
                "that the LP writer does not know how to process:\n\t%s"
                % (
                    model.name,
                    "\n\t".join(
                        "%s:\n\t\t%s" % (k, "\n\t\t".join(map(attrgetter('name'), v)))
                        for k, v in unknown.items()
                    ),
                )
            )

        ONE_VAR_CONSTANT = Var(name='ONE_VAR_CONSTANT', bounds=(1, 1))
        ONE_VAR_CONSTANT.construct()

        self.var_map = {id(ONE_VAR_CONSTANT): ONE_VAR_CONSTANT}
        initialize_var_map_from_column_order(model, self.config, self.var_map)
        self.var_order = {_id: i for i, _id in enumerate(self.var_map)}
        self.var_recorder = OrderedVarRecorder(self.var_map, self.var_order, sorter)

        _qp = self.config.allow_quadratic_objective
        _qc = self.config.allow_quadratic_constraint
        objective_visitor = (QuadraticRepnVisitor if _qp else LinearRepnVisitor)(
            {}, var_recorder=self.var_recorder
        )
        constraint_visitor = (QuadraticRepnVisitor if _qc else LinearRepnVisitor)(
            objective_visitor.subexpression_cache if _qp == _qc else {},
            var_recorder=self.var_recorder,
        )

        timer.toc('Initialized column order', level=logging.DEBUG)

        # We don't export any suffix information to the LP file
        #
        if component_map[Suffix]:
            suffixesByName = {}
            for block in component_map[Suffix]:
                for suffix in block.component_objects(
                    Suffix, active=True, descend_into=False, sort=sorter
                ):
                    if not suffix.export_enabled() or not suffix:
                        continue
                    name = suffix.local_name
                    if name in suffixesByName:
                        suffixesByName[name].append(suffix)
                    else:
                        suffixesByName[name] = [suffix]
            for name, suffixes in suffixesByName.items():
                n = len(suffixes)
                plural = 's' if n > 1 else ''
                logger.warning(
                    f"EXPORT Suffix '{name}' found on {n} block{plural}:\n    "
                    + "\n    ".join(s.name for s in suffixes)
                    + "\nLP writer cannot export suffixes to LP files.  Skipping."
                )

        ostream.write(f"\\* Source Pyomo model name={model.name} *\\\n\n")

        #
        # Process objective
        #
        if not component_map[Objective]:
            objectives = [Objective(expr=1)]
            objectives[0].construct()
        else:
            objectives = []
            for blk in component_map[Objective]:
                objectives.extend(
                    blk.component_data_objects(
                        Objective, active=True, descend_into=False, sort=sorter
                    )
                )
        if len(objectives) > 1:
            raise ValueError(
                "More than one active objective defined for input model '%s'; "
                "Cannot write legal LP file\nObjectives: %s"
                % (model.name, ' '.join(obj.name for obj in objectives))
            )

        obj = objectives[0]
        ostream.write(
            ("min \n%s:\n" if obj.sense == minimize else "max \n%s:\n")
            % (getSymbol(obj, labeler),)
        )
        repn = objective_visitor.walk_expression(obj.expr)
        if repn.nonlinear is not None:
            raise ValueError(
                f"Model objective ({obj.name}) contains nonlinear terms that "
                "cannot be written to LP format"
            )
        if repn.constant or not (repn.linear or getattr(repn, 'quadratic', None)):
            # Older versions of CPLEX (including 12.6) and all versions
            # of GLPK (through 5.0) do not support constants in the
            # objective in LP format.  To avoid painful bookkeeping, we
            # introduce the following "variable", constrained to the
            # value 1.
            #
            # In addition, most solvers do no tolerate an empty
            # objective, this will ensure we at least write out
            # 0*ONE_VAR_CONSTANT.
            repn.linear[id(ONE_VAR_CONSTANT)] = repn.constant
            repn.constant = 0
        self.write_expression(ostream, repn, True)
        aliasSymbol(obj, '__default_objective__')
        if with_debug_timing:
            timer.toc('Objective %s', obj, level=logging.DEBUG)

        ostream.write("\ns.t.\n")

        #
        # Tabulate constraints
        #
        skip_trivial_constraints = self.config.skip_trivial_constraints
        have_nontrivial = False
        last_parent = None
        for con in ordered_active_constraints(model, self.config):
            if with_debug_timing and con.parent_component() is not last_parent:
                timer.toc('Constraint %s', last_parent, level=logging.DEBUG)
                last_parent = con.parent_component()
            # Note: Constraint.to_bounded_expression(evaluate_bounds=True)
            # guarantee a return value that is either a (finite)
            # native_numeric_type, or None
            lb, body, ub = con.to_bounded_expression(True)

            if lb is None and ub is None:
                # Note: you *cannot* output trivial (unbounded)
                # constraints in LP format.  I suppose we could add a
                # slack variable if skip_trivial_constraints is False,
                # but that seems rather silly.
                continue
            repn = constraint_visitor.walk_expression(body)
            if repn.nonlinear is not None:
                raise ValueError(
                    f"Model constraint ({con.name}) contains nonlinear terms that "
                    "cannot be written to LP format"
                )

            # Pull out the constant: we will move it to the bounds
            offset = repn.constant
            repn.constant = 0

            if repn.linear or getattr(repn, 'quadratic', None):
                have_nontrivial = True
            else:
                if (
                    skip_trivial_constraints
                    and (lb is None or lb <= offset)
                    and (ub is None or ub >= offset)
                ):
                    continue
                # This is a trivially infeasible model.  We could raise
                # an exception, or we could allow the solver to return
                # infeasible.  There are fewer logic paths (in
                # particular related to mapping solver result status) if
                # we just defer to the solver.
                #
                # Add a dummy (fixed) variable to the constraint,
                # because some solvers (including versions of GLPK)
                # cannot parse an LP file without a variable on the left
                # hand side.
                repn.linear[id(ONE_VAR_CONSTANT)] = 0

            symbol = labeler(con)
            if lb is not None:
                if ub is None:
                    label = f'c_l_{symbol}_'
                    addSymbol(con, label)
                    ostream.write(f'\n{label}:\n')
                    self.write_expression(ostream, repn, False)
                    ostream.write(f'>= {(lb - offset)!s}\n')
                elif lb == ub:
                    label = f'c_e_{symbol}_'
                    addSymbol(con, label)
                    ostream.write(f'\n{label}:\n')
                    self.write_expression(ostream, repn, False)
                    ostream.write(f'= {(lb - offset)!s}\n')
                else:
                    # We will need the constraint body twice.  Generate
                    # in a buffer so we only have to do that once.
                    buf = StringIO()
                    self.write_expression(buf, repn, False)
                    buf = buf.getvalue()
                    #
                    label = f'r_l_{symbol}_'
                    addSymbol(con, label)
                    ostream.write(f'\n{label}:\n')
                    ostream.write(buf)
                    ostream.write(f'>= {(lb - offset)!s}\n')
                    label = f'r_u_{symbol}_'
                    aliasSymbol(con, label)
                    ostream.write(f'\n{label}:\n')
                    ostream.write(buf)
                    ostream.write(f'<= {(ub - offset)!s}\n')
            elif ub is not None:
                label = f'c_u_{symbol}_'
                addSymbol(con, label)
                ostream.write(f'\n{label}:\n')
                self.write_expression(ostream, repn, False)
                ostream.write(f'<= {(ub - offset)!s}\n')

        if with_debug_timing:
            # report the last constraint
            timer.toc('Constraint %s', last_parent, level=logging.DEBUG)
        if not have_nontrivial:
            # Some solvers (notably CBC through at least 2.10.4) will
            # return a nonzero return code when the model has no
            # constraints.  To work around the original Pyomo solver
            # hierarchy (where the return code was processed in the base
            # class), we will add a dummy constraint here.
            repn = constraint_visitor.Result()  # walk_expression(ONE_VAR_CONSTANT)
            repn.linear[id(ONE_VAR_CONSTANT)] = 1
            ostream.write(f'\nc_e_ONE_VAR_CONSTANT:\n')
            self.write_expression(ostream, repn, False)
            ostream.write(f'= 1\n')

        ostream.write("\nbounds")

        # Track the number of integer and binary variables, so you can
        # output their status later.
        integer_vars = []
        binary_vars = []
        getSymbolByObjectID = self.symbol_map.byObject.get
        for vid, v in self.var_map.items():
            # Some variables in the var_map may not actually have been
            # written out to the LP file (e.g., added from col_order, or
            # multiplied by 0 in the expressions).  Check to see that
            # the variable is in the symbol_map before outputting.
            v_symbol = getSymbolByObjectID(vid, None)
            if not v_symbol:
                continue
            if v.is_binary():
                binary_vars.append(v_symbol)
            elif v.is_integer():
                integer_vars.append(v_symbol)

            # Note: Var.bounds guarantees the values are either (finite)
            # native_numeric_types or None
            lb, ub = v.bounds
            lb = '-inf' if lb is None else str(lb)
            ub = '+inf' if ub is None else str(ub)
            ostream.write(f"\n   {lb} <= {v_symbol} <= {ub}")

        if integer_vars:
            ostream.write("\ngeneral\n  ")
            ostream.write("\n  ".join(integer_vars))

        if binary_vars:
            ostream.write("\nbinary\n  ")
            ostream.write("\n  ".join(binary_vars))

        timer.toc("Wrote variable bounds and domains", level=logging.DEBUG)

        #
        # Tabulate SOS constraints
        #
        if component_map[SOSConstraint]:
            sos = []
            for blk in component_map[SOSConstraint]:
                sos.extend(
                    blk.component_data_objects(
                        SOSConstraint, active=True, descend_into=False, sort=sorter
                    )
                )
            if self.config.row_order:
                row_map = row_order2row_map(self.config)
                _n = len(row_map)
                sos.sort(key=lambda x: row_map.get(id(x), _n))

            ostream.write("\nSOS\n")
            for soscon in sos:
                ostream.write(f'\n{getSymbol(soscon)}: S{soscon.level}::\n')
                for v, w in getattr(soscon, 'get_items', soscon.items)():
                    if w.__class__ not in int_float:
                        w = float(w)
                    ostream.write(f"  {getSymbol(v)}:{w!s}\n")

        ostream.write("\nend\n")

        info = LPWriterInfo(self.symbol_map)
        timer.toc("Generated LP representation", delta=False)
        return info

    def write_expression(self, ostream, expr, is_objective):
        assert not expr.constant
        getSymbol = self.symbol_map.getSymbol
        getVarOrder = self.var_order.__getitem__
        getVar = self.var_map.__getitem__

        if expr.linear:
            for vid, coef in sorted(
                expr.linear.items(), key=lambda x: getVarOrder(x[0])
            ):
                if coef < 0:
                    ostream.write(f'{coef!s} {getSymbol(getVar(vid))}\n')
                else:
                    ostream.write(f'+{coef!s} {getSymbol(getVar(vid))}\n')

        quadratic = getattr(expr, 'quadratic', None)
        if quadratic:

            def _normalize_constraint(data):
                (vid1, vid2), coef = data
                c1 = getVarOrder(vid1)
                c2 = getVarOrder(vid2)
                if c2 < c1:
                    col = c2, c1
                    sym = f' {getSymbol(getVar(vid2))} * {getSymbol(getVar(vid1))}\n'
                elif c1 == c2:
                    col = c1, c1
                    sym = f' {getSymbol(getVar(vid2))} ^ 2\n'
                else:
                    col = c1, c2
                    sym = f' {getSymbol(getVar(vid1))} * {getSymbol(getVar(vid2))}\n'
                if coef < 0:
                    return col, str(coef) + sym
                else:
                    return col, f'+{coef!s}{sym}'

            if is_objective:
                #
                # Times 2 because LP format requires /2 for all the
                # quadratic terms /of the objective only/.  Discovered
                # the last bit through trial and error.
                # Ref: ILog CPlex 8.0 User's Manual, p197.
                #
                def _normalize_objective(data):
                    vids, coef = data
                    return _normalize_constraint((vids, 2 * coef))

                _normalize = _normalize_objective
            else:
                _normalize = _normalize_constraint

            ostream.write('+ [\n')
            quadratic = sorted(map(_normalize, quadratic.items()), key=itemgetter(0))
            ostream.write(''.join(map(itemgetter(1), quadratic)))
            if is_objective:
                ostream.write("] / 2\n")
            else:
                ostream.write("]\n")



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from functools import wraps

from pyomo.common.autoslots import AutoSlots
from pyomo.common.collections import ComponentMap, DefaultComponentMap
from pyomo.common.log import is_debug_set
from pyomo.common.modeling import unique_component_name

from pyomo.core.base import Transformation, TransformationFactory
from pyomo.core.base.external import ExternalFunction
from pyomo.core import (
    Any,
    Block,
    BooleanVar,
    Connector,
    Constraint,
    Expression,
    NonNegativeIntegers,
    Param,
    RangeSet,
    Reference,
    Set,
    SetOf,
    Suffix,
    Var,
)
from pyomo.gdp import Disjunct, Disjunction, GDP_Error
from pyomo.gdp.transformed_disjunct import _TransformedDisjunct
from pyomo.gdp.util import (
    get_gdp_tree,
    get_src_constraint,
    get_src_disjunct,
    get_src_disjunction,
    get_transformed_constraints,
    _warn_for_active_disjunct,
)
from pyomo.network import Port

from weakref import ref as weakref_ref


class _GDPTransformationData(AutoSlots.Mixin):
    __slots__ = ('src_constraint', 'transformed_constraints')

    def __init__(self):
        self.src_constraint = ComponentMap()
        self.transformed_constraints = DefaultComponentMap(list)


Block.register_private_data_initializer(_GDPTransformationData, scope='pyomo.gdp')


class GDP_to_MIP_Transformation(Transformation):
    """
    Base class for transformations from GDP to MIP
    """

    def __init__(self, logger):
        """Initialize transformation object."""
        super(GDP_to_MIP_Transformation, self).__init__()
        self.logger = logger
        self.handlers = {
            Constraint: self._transform_constraint,
            Var: False,  # Note that if a Var appears on a Disjunct, we
            # still treat its bounds as global. If the
            # intent is for its bounds to be on the
            # disjunct, it should be declared with no bounds
            # and the bounds should be set in constraints on
            # the Disjunct.
            BooleanVar: False,
            Connector: False,
            Expression: False,
            Suffix: False,
            Param: False,
            Set: False,
            SetOf: False,
            RangeSet: False,
            Disjunction: False,  # In BigM, it's impossible to encounter an active
            # Disjunction because preprocessing would have
            # put it before its parent Disjunct in the order
            # of transformation. In hull, we intentionally
            # pass over active Disjunctions that are on
            # Disjuncts because we know they are in the list
            # of objects to transform after preprocessing, so
            # they will be transformed later.
            Disjunct: self._warn_for_active_disjunct,
            Block: False,
            ExternalFunction: False,
            Port: False,  # not Arcs, because those are deactivated after
            # the network.expand_arcs transformation
        }
        self._generate_debug_messages = False
        self._transformation_blocks = {}
        self._algebraic_constraints = {}

    def _restore_state(self):
        self._transformation_blocks.clear()
        self._algebraic_constraints.clear()
        if hasattr(self, '_config'):
            del self._config

    def _process_arguments(self, instance, **kwds):
        if not instance.ctype in (Block, Disjunct):
            raise GDP_Error(
                "Transformation called on %s of type %s. 'instance' "
                "must be a ConcreteModel, Block, or Disjunct (in "
                "the case of nested disjunctions)." % (instance.name, instance.ctype)
            )

        self._config = self.CONFIG(kwds.pop('options', {}))
        self._config.set_value(kwds)
        self._generate_debug_messages = is_debug_set(self.logger)

    def _transform_logical_constraints(self, instance, targets):
        # transform any logical constraints that might be anywhere on the stuff
        # we're about to transform. We do this before we preprocess targets
        # because we will likely create more disjunctive components that will
        # need transformation.
        disj_targets = []
        for t in targets:
            disj_datas = t.values() if t.is_indexed() else [t]
            if t.ctype is Disjunct:
                disj_targets.extend(disj_datas)
            if t.ctype is Disjunction:
                disj_targets.extend(
                    [d for disjunction in disj_datas for d in disjunction.disjuncts]
                )
        TransformationFactory('contrib.logical_to_disjunctive').apply_to(
            instance,
            targets=[blk for blk in targets if blk.ctype is Block] + disj_targets,
        )

    def _filter_targets(self, instance):
        targets = self._config.targets
        if targets is None:
            targets = (instance,)

        # FIXME: For historical reasons, Hull would silently skip
        # any targets that were explicitly deactivated.  This
        # preserves that behavior (although adds a warning).  We
        # should revisit that design decision and probably remove
        # this filter, as it is slightly ambiguous as to what it
        # means for the target to be deactivated: is it just the
        # target itself [historical implementation] or any block in
        # the hierarchy?
        def _filter_inactive(targets):
            for t in targets:
                if not t.active:
                    self.logger.warning(
                        f'GDP.{self.transformation_name} transformation passed '
                        f'a deactivated target ({t.name}). Skipping.'
                    )
                else:
                    yield t

        return list(_filter_inactive(targets))

    def _get_gdp_tree_from_targets(self, instance, targets):
        knownBlocks = {}
        # we need to preprocess targets to make sure that if there are any
        # disjunctions in targets that they appear before disjunctions that are
        # contained in their disjuncts. That is, in hull, we will transform from
        # root to leaf in order to avoid have to modify transformed constraints
        # more than once: It is most efficient to build nested transformed
        # constraints when we already have the disaggregated variables of the
        # parent disjunct.
        return get_gdp_tree(targets, instance)

    def _add_transformation_block(self, to_block):
        if to_block in self._transformation_blocks:
            return self._transformation_blocks[to_block], False

        # make a transformation block on to_block to put transformed disjuncts
        # on
        transBlockName = unique_component_name(
            to_block, '_pyomo_gdp_%s_reformulation' % self.transformation_name
        )
        self._transformation_blocks[to_block] = transBlock = Block()
        to_block.add_component(transBlockName, transBlock)
        transBlock.relaxedDisjuncts = _TransformedDisjunct(NonNegativeIntegers)

        return transBlock, True

    def _add_xor_constraint(self, disjunction, transBlock):
        # Put the disjunction constraint on the transformation block and
        # determine whether it is an OR or XOR constraint.
        # We never do this for just a DisjunctionData because we need to know
        # about the index set of its parent component (so that we can make the
        # index of this constraint match). So if we called this on a
        # DisjunctionData, we did something wrong.

        # first check if the constraint already exists
        if disjunction in self._algebraic_constraints:
            return self._algebraic_constraints[disjunction]

        # add the XOR (or OR) constraints to parent block (with unique name)
        # It's indexed if this is an IndexedDisjunction, not otherwise
        if disjunction.is_indexed():
            orC = Constraint(Any)
        else:
            orC = Constraint()
        orCname = unique_component_name(
            transBlock, disjunction.getname(fully_qualified=False) + '_xor'
        )
        transBlock.add_component(orCname, orC)
        self._algebraic_constraints[disjunction] = orC

        return orC

    def _setup_transform_disjunctionData(self, obj, root_disjunct):
        # Just because it's unlikely this is what someone meant to do...
        if len(obj.disjuncts) == 0:
            raise GDP_Error(
                "Disjunction '%s' is empty. This is "
                "likely indicative of a modeling error." % obj.name
            )

        # We always need to create or fetch a transformation block on the parent block.
        trans_block, new_block = self._add_transformation_block(obj.parent_block())
        # This is where we put exactly_one/or constraint
        algebraic_constraint = self._add_xor_constraint(
            obj.parent_component(), trans_block
        )

        # If requested, create or fetch the transformation block above the
        # nested hierarchy
        if root_disjunct is not None:
            # We want to put some transformed things on the root Disjunct's
            # parent's block so that they do not get re-transformed. (Note this
            # is never true for hull, but it calls this method with
            # root_disjunct=None. BigM can't put the exactly-one constraint up
            # here, but it can put everything else.)
            trans_block, new_block = self._add_transformation_block(
                root_disjunct.parent_block()
            )

        return trans_block, algebraic_constraint

    def _get_disjunct_transformation_block(self, disjunct, transBlock):
        if disjunct.transformation_block is not None:
            return disjunct.transformation_block

        # create a relaxation block for this disjunct
        relaxedDisjuncts = transBlock.relaxedDisjuncts
        relaxationBlock = relaxedDisjuncts[len(relaxedDisjuncts)]

        relaxationBlock.transformedConstraints = Constraint(Any)
        relaxationBlock.localVarReferences = Block()

        # add mappings to source disjunct (so we'll know we've relaxed)
        disjunct._transformation_block = weakref_ref(relaxationBlock)
        relaxationBlock._src_disjunct = weakref_ref(disjunct)

        return relaxationBlock

    def _transform_block_components(self, block, disjunct, *args):
        # Find all the variables declared here (including the indicator_var) and
        # add a reference on the transformation block so these will be
        # accessible when the Disjunct is deactivated. Note that in hull, we do
        # this after we have moved up the transformation blocks for nested
        # disjunctions, so that we don't have duplicate references.
        varRefBlock = disjunct._transformation_block().localVarReferences
        for v in block.component_objects(Var, descend_into=Block, active=None):
            varRefBlock.add_component(
                unique_component_name(varRefBlock, v.getname(fully_qualified=False)),
                Reference(v),
            )

        # Now look through the component map of block and transform everything
        # we have a handler for. Yell if we don't know how to handle it. (Note
        # that because we only iterate through active components, this means
        # non-ActiveComponent types cannot have handlers.)
        for obj in block.component_objects(active=True, descend_into=Block):
            handler = self.handlers.get(obj.ctype, None)
            if not handler:
                if handler is None:
                    raise GDP_Error(
                        "No %s transformation handler registered "
                        "for modeling components of type %s. If your "
                        "disjuncts contain non-GDP Pyomo components that "
                        "require transformation, please transform them first."
                        % (self.transformation_name, obj.ctype)
                    )
                continue
            # obj is what we are transforming, we pass disjunct
            # through so that we will have access to the indicator
            # variables down the line.
            handler(obj, disjunct, *args)

    def _transform_constraint(self, obj, disjunct, *args):
        raise NotImplementedError(
            "Class %s failed to implement '_transform_constraint'" % self.__class__
        )

    def _warn_for_active_disjunct(self, innerdisjunct, outerdisjunct, *args):
        _warn_for_active_disjunct(innerdisjunct, outerdisjunct)

    # These are all functions to retrieve transformed components from
    # original ones and vice versa.

    @wraps(get_src_disjunct)
    def get_src_disjunct(self, transBlock):
        return get_src_disjunct(transBlock)

    @wraps(get_src_disjunction)
    def get_src_disjunction(self, xor_constraint):
        return get_src_disjunction(xor_constraint)

    @wraps(get_src_constraint)
    def get_src_constraint(self, transformedConstraint):
        return get_src_constraint(transformedConstraint)

    @wraps(get_transformed_constraints)
    def get_transformed_constraints(self, srcConstraint):
        return get_transformed_constraints(srcConstraint)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Big-M Generalized Disjunctive Programming transformation module."""

import logging

from pyomo.common.autoslots import AutoSlots
from pyomo.common.collections import ComponentMap
from pyomo.common.config import ConfigDict, ConfigValue
from pyomo.common.gc_manager import PauseGC
from pyomo.common.modeling import unique_component_name
from pyomo.common.deprecation import deprecated, deprecation_warning
from pyomo.contrib.cp.transform.logical_to_disjunctive_program import (
    LogicalToDisjunctive,
)
from pyomo.core import (
    Block,
    BooleanVar,
    Connector,
    Constraint,
    Param,
    Set,
    SetOf,
    Var,
    Expression,
    SortComponents,
    TraversalStrategy,
    value,
    RangeSet,
    NonNegativeIntegers,
    Binary,
    Any,
)
from pyomo.core.base import TransformationFactory, Reference
import pyomo.core.expr as EXPR
from pyomo.gdp import Disjunct, Disjunction, GDP_Error
from pyomo.gdp.plugins.bigm_mixin import (
    _BigM_MixIn,
    _get_bigM_suffix_list,
    _warn_for_unused_bigM_args,
)
from pyomo.gdp.plugins.gdp_to_mip_transformation import GDP_to_MIP_Transformation
from pyomo.gdp.transformed_disjunct import _TransformedDisjunct
from pyomo.gdp.util import is_child_of, _get_constraint_transBlock, _to_dict
from pyomo.core.util import target_list
from pyomo.network import Port
from pyomo.repn import generate_standard_repn
from weakref import ref as weakref_ref, ReferenceType

logger = logging.getLogger('pyomo.gdp.bigm')


class _BigMData(AutoSlots.Mixin):
    __slots__ = ('bigm_src',)

    def __init__(self):
        # we will keep a map of constraints (hashable, ha!) to a tuple to
        # indicate what their M value is and where it came from, of the form:
        # ((lower_value, lower_source, lower_key), (upper_value, upper_source,
        # upper_key)), where the first tuple is the information for the lower M,
        # the second tuple is the info for the upper M, source is the Suffix or
        # argument dictionary and None if the value was calculated, and key is
        # the key in the Suffix or argument dictionary, and None if it was
        # calculated. (Note that it is possible the lower or upper is
        # user-specified and the other is not, hence the need to store
        # information for both.)
        self.bigm_src = {}


Block.register_private_data_initializer(_BigMData)


@TransformationFactory.register(
    'gdp.bigm', doc="Relax disjunctive model using big-M terms."
)
class BigM_Transformation(GDP_to_MIP_Transformation, _BigM_MixIn):
    """Relax disjunctive model using big-M terms.

    Relaxes a disjunctive model into an algebraic model by adding Big-M
    terms to all disjunctive constraints.

    This transformation accepts the following keyword arguments:
        bigM: A user-specified value (or dict) of M values to use (see below)
        targets: the targets to transform [default: the instance]

    M values are determined as follows:
       1. if the constraint appears in the bigM argument dict
       2. if the constraint parent_component appears in the bigM argument dict
       3. if any block which is an ancestor to the constraint appears in
          the bigM argument dict
       4. if 'None' is in the bigM argument dict
       5. if the constraint or the constraint parent_component appear in
          a BigM Suffix attached to any parent_block() beginning with the
          constraint's parent_block and moving up to the root model.
       6. if None appears in a BigM Suffix attached to any
          parent_block() between the constraint and the root model.
       7. if the constraint is linear, estimate M using the variable bounds

    M values may be a single value or a 2-tuple specifying the M for the
    lower bound and the upper bound of the constraint body.

    Specifying "bigM=N" is automatically mapped to "bigM={None: N}".

    The transformation will create a new Block with a unique
    name beginning "_pyomo_gdp_bigm_reformulation".  That Block will
    contain an indexed Block named "relaxedDisjuncts", which will hold
    the relaxed disjuncts.  This block is indexed by an integer
    indicating the order in which the disjuncts were relaxed. All
    transformed Disjuncts will have a pointer to the block their transformed
    constraints are on, and all transformed Disjunctions will have a
    pointer to the corresponding 'Or' or 'ExactlyOne' constraint.

    """

    CONFIG = ConfigDict("gdp.bigm")
    CONFIG.declare(
        'targets',
        ConfigValue(
            default=None,
            domain=target_list,
            description="target or list of targets that will be relaxed",
            doc="""

        This specifies the list of components to relax. If None (default), the
        entire model is transformed. Note that if the transformation is done out
        of place, the list of targets should be attached to the model before it
        is cloned, and the list will specify the targets on the cloned
        instance.""",
        ),
    )
    CONFIG.declare(
        'bigM',
        ConfigValue(
            default=None,
            domain=_to_dict,
            description="Big-M value used for constraint relaxation",
            doc="""

        A user-specified value, dict, or ComponentMap of M values that override
        M-values found through model Suffixes or that would otherwise be
        calculated using variable domains.""",
        ),
    )
    CONFIG.declare(
        'assume_fixed_vars_permanent',
        ConfigValue(
            default=False,
            domain=bool,
            description="Boolean indicating whether or not to transform so that "
            "the transformed model will still be valid when fixed Vars are "
            "unfixed.",
            doc="""
        This is only relevant when the transformation will be estimating values
        for M. If True, the transformation will calculate M values assuming that
        fixed variables will always be fixed to their current values. This means
        that if a fixed variable is unfixed after transformation, the
        transformed model is potentially no longer valid. By default, the
        transformation will assume fixed variables could be unfixed in the
        future and will use their bounds to calculate the M value rather than
        their value. Note that this could make for a weaker LP relaxation
        while the variables remain fixed.
        """,
        ),
    )
    transformation_name = 'bigm'

    def __init__(self):
        super().__init__(logger)
        self._set_up_expr_bound_visitor()

    def _apply_to(self, instance, **kwds):
        self.used_args = ComponentMap()  # If everything was sure to go well,
        # this could be a dictionary. But if
        # someone messes up and gives us a Var
        # as a key in bigMargs, I need the error
        # not to be when I try to put it into
        # this map!
        with PauseGC():
            try:
                self._apply_to_impl(instance, **kwds)
            finally:
                self._restore_state()
                self.used_args.clear()
                self._expr_bound_visitor.leaf_bounds.clear()
                self._expr_bound_visitor.use_fixed_var_values_as_bounds = False

    def _apply_to_impl(self, instance, **kwds):
        self._process_arguments(instance, **kwds)
        if self._config.assume_fixed_vars_permanent:
            self._expr_bound_visitor.use_fixed_var_values_as_bounds = True

        # filter out inactive targets and handle case where targets aren't
        # specified.
        targets = self._filter_targets(instance)
        # transform logical constraints based on targets
        self._transform_logical_constraints(instance, targets)
        # we need to preprocess targets to make sure that if there are any
        # disjunctions in targets that their disjuncts appear before them in
        # the list.
        gdp_tree = self._get_gdp_tree_from_targets(instance, targets)
        preprocessed_targets = gdp_tree.reverse_topological_sort()

        bigM = self._config.bigM
        for t in preprocessed_targets:
            if t.ctype is Disjunction:
                self._transform_disjunctionData(t, t.index(), bigM, gdp_tree)

        # issue warnings about anything that was in the bigM args dict that we
        # didn't use
        _warn_for_unused_bigM_args(bigM, self.used_args, logger)

    def _transform_disjunctionData(self, obj, index, bigM, gdp_tree):
        parent_disjunct = gdp_tree.parent(obj)
        root_disjunct = gdp_tree.root_disjunct(obj)
        transBlock, xorConstraint = self._setup_transform_disjunctionData(
            obj, root_disjunct
        )

        # add or (or xor) constraint
        or_expr = 0
        for disjunct in obj.disjuncts:
            or_expr += disjunct.binary_indicator_var
            self._transform_disjunct(disjunct, bigM, transBlock, gdp_tree)

        if obj.xor:
            xorConstraint[index] = or_expr == 1
        else:
            xorConstraint[index] = or_expr >= 1
        # Mark the DisjunctionData as transformed by mapping it to its XOR
        # constraint.
        obj._algebraic_constraint = weakref_ref(xorConstraint[index])

        # and deactivate for the writers
        obj.deactivate()

    def _transform_disjunct(self, obj, bigM, transBlock, gdp_tree):
        # We're not using the preprocessed list here, so this could be
        # inactive. We've already done the error checking in preprocessing, so
        # we just skip it here.
        if not obj.active:
            return

        suffix_list = _get_bigM_suffix_list(obj)
        arg_list = self._get_bigM_arg_list(bigM, obj)

        relaxationBlock = self._get_disjunct_transformation_block(obj, transBlock)

        indicator_expression = 0
        node = obj
        while node is not None:
            indicator_expression += 1 - node.binary_indicator_var
            node = gdp_tree.parent_disjunct(node)

        # This is crazy, but if the disjunction has been previously
        # relaxed, the disjunct *could* be deactivated.  This is a big
        # deal for Hull, as it uses the component_objects /
        # component_data_objects generators.  For BigM, that is OK,
        # because we never use those generators with active=True.  I am
        # only noting it here for the future when someone (me?) is
        # comparing the two relaxations.
        #
        # Transform each component within this disjunct
        self._transform_block_components(
            obj, obj, bigM, arg_list, suffix_list, indicator_expression
        )

        # deactivate disjunct to keep the writers happy
        obj._deactivate_without_fixing_indicator()

    def _transform_constraint(
        self,
        obj,
        disjunct,
        bigMargs,
        arg_list,
        disjunct_suffix_list,
        indicator_expression,
    ):
        # add constraint to the transformation block, we'll transform it there.
        transBlock = disjunct._transformation_block()
        bigm_src = transBlock.private_data().bigm_src
        constraint_map = transBlock.private_data('pyomo.gdp')

        disjunctionRelaxationBlock = transBlock.parent_block()

        # We will make indexes from ({obj.local_name} x obj.index_set() x ['lb',
        # 'ub']), but don't bother construct that set here, as taking Cartesian
        # products is kind of expensive (and redundant since we have the
        # original model)
        newConstraint = transBlock.transformedConstraints

        for i in sorted(obj.keys()):
            c = obj[i]
            if not c.active:
                continue

            lower = (None, None, None)
            upper = (None, None, None)

            # first, we see if an M value was specified in the arguments.
            # (This returns None if not)
            lower, upper = self._get_M_from_args(c, bigMargs, arg_list, lower, upper)
            M = (lower[0], upper[0])

            if self._generate_debug_messages:
                logger.debug(
                    "GDP(BigM): The value for M for constraint '%s' "
                    "from the BigM argument is %s." % (c.name, str(M))
                )

            # if we didn't get something we need from args, try suffixes:
            if (M[0] is None and c.lower is not None) or (
                M[1] is None and c.upper is not None
            ):
                # first get anything parent to c but below disjunct
                suffix_list = _get_bigM_suffix_list(
                    c.parent_block(), stopping_block=disjunct
                )
                # prepend that to what we already collected for the disjunct.
                suffix_list.extend(disjunct_suffix_list)
                lower, upper = self._update_M_from_suffixes(
                    c, suffix_list, lower, upper
                )
                M = (lower[0], upper[0])

            if self._generate_debug_messages:
                logger.debug(
                    "GDP(BigM): The value for M for constraint '%s' "
                    "after checking suffixes is %s." % (c.name, str(M))
                )

            if c.lower is not None and M[0] is None:
                M = (self._estimate_M(c.body, c)[0] - c.lower, M[1])
                lower = (M[0], None, None)
            if c.upper is not None and M[1] is None:
                M = (M[0], self._estimate_M(c.body, c)[1] - c.upper)
                upper = (M[1], None, None)

            if self._generate_debug_messages:
                logger.debug(
                    "GDP(BigM): The value for M for constraint '%s' "
                    "after estimating (if needed) is %s." % (c.name, str(M))
                )

            # save the source information
            bigm_src[c] = (lower, upper)

            self._add_constraint_expressions(
                c,
                i,
                M,
                disjunct.binary_indicator_var,
                newConstraint,
                constraint_map,
                indicator_expression=indicator_expression,
            )

            # deactivate because we relaxed
            c.deactivate()

    def _update_M_from_suffixes(self, constraint, suffix_list, lower, upper):
        # It's possible we found half the answer in args, but we are still
        # looking for half the answer.
        need_lower = constraint.lower is not None and lower[0] is None
        need_upper = constraint.upper is not None and upper[0] is None
        M = None
        # first we check if the constraint or its parent is a key in any of the
        # suffix lists
        for bigm in suffix_list:
            if constraint in bigm:
                M = bigm[constraint]
                lower, upper, need_lower, need_upper = self._process_M_value(
                    M,
                    lower,
                    upper,
                    need_lower,
                    need_upper,
                    bigm,
                    constraint,
                    constraint,
                )
                if not need_lower and not need_upper:
                    return lower, upper

            # if c is indexed, check for the parent component
            if constraint.parent_component() in bigm:
                parent = constraint.parent_component()
                M = bigm[parent]
                lower, upper, need_lower, need_upper = self._process_M_value(
                    M, lower, upper, need_lower, need_upper, bigm, parent, constraint
                )
                if not need_lower and not need_upper:
                    return lower, upper

        # if we didn't get an M that way, traverse upwards through the blocks
        # and see if None has a value on any of them.
        if M is None:
            for bigm in suffix_list:
                if None in bigm:
                    M = bigm[None]
                    lower, upper, need_lower, need_upper = self._process_M_value(
                        M, lower, upper, need_lower, need_upper, bigm, None, constraint
                    )
                if not need_lower and not need_upper:
                    return lower, upper
        return lower, upper

    @deprecated(
        "The get_m_value_src function is deprecated. Use "
        "the get_M_value_src function if you need source "
        "information or the get_M_value function if you "
        "only need values.",
        version='5.7.1',
    )
    def get_m_value_src(self, constraint):
        transBlock = _get_constraint_transBlock(constraint)
        (lower_val, lower_source, lower_key), (upper_val, upper_source, upper_key) = (
            transBlock.private_data().bigm_src[constraint]
        )

        if (
            constraint.lower is not None
            and constraint.upper is not None
            and (not lower_source is upper_source or not lower_key is upper_key)
        ):
            raise GDP_Error(
                "This is why this method is deprecated: The lower "
                "and upper M values for constraint %s came from "
                "different sources, please use the get_M_value_src "
                "method." % constraint.name
            )
        # if source and key are equal for the two, this is representable in the
        # old format.
        if constraint.lower is not None and lower_source is not None:
            return (lower_source, lower_key)
        if constraint.upper is not None and upper_source is not None:
            return (upper_source, upper_key)
        # else it was calculated:
        return (lower_val, upper_val)

    def get_M_value_src(self, constraint):
        """Return a tuple indicating how the M value used to transform
        constraint was specified. (In particular, this can be used to
        verify which BigM Suffixes were actually necessary to the
        transformation.)

        Return is of the form: ((lower_M_val, lower_M_source, lower_M_key),
                                (upper_M_val, upper_M_source, upper_M_key))

        If the constraint does not have a lower bound (or an upper bound),
        the first (second) element will be (None, None, None). Note that if
        a constraint is of the form a <= expr <= b or is an equality constraint,
        it is not necessarily true that the source of lower_M and upper_M
        are the same.

        If the M value came from an arg, source is the  dictionary itself and
        key is the key in that dictionary which gave us the M value.

        If the M value came from a Suffix, source is the BigM suffix used and
        key is the key in that Suffix.

        If the transformation calculated the value, both source and key are
        None.

        Parameters
        ----------
        constraint: Constraint, which must be in the subtree of a transformed
                    Disjunct
        """
        transBlock = _get_constraint_transBlock(constraint)
        # This is a KeyError if it fails, but it is also my fault if it
        # fails... (That is, it's a bug in the mapping.)
        return transBlock.private_data().bigm_src[constraint]

    def get_M_value(self, constraint):
        """Returns the M values used to transform constraint. Return is a tuple:
        (lower_M_value, upper_M_value). Either can be None if constraint does
        not have a lower or upper bound, respectively.

        Parameters
        ----------
        constraint: Constraint, which must be in the subtree of a transformed
                    Disjunct
        """
        transBlock = _get_constraint_transBlock(constraint)
        # This is a KeyError if it fails, but it is also my fault if it
        # fails... (That is, it's a bug in the mapping.)
        lower, upper = transBlock.private_data().bigm_src[constraint]
        return (lower[0], upper[0])

    def get_all_M_values_by_constraint(self, model):
        """Returns a dictionary mapping each constraint to a tuple:
        (lower_M_value, upper_M_value), where either can be None if the
        constraint does not have a lower or upper bound (respectively).

        Parameters
        ----------
        model: A GDP model that has been transformed with BigM
        """
        m_values = {}
        for disj in model.component_data_objects(
            Disjunct, active=None, descend_into=(Block, Disjunct)
        ):
            transBlock = disj.transformation_block
            # First check if it was transformed at all.
            if transBlock is not None:
                # If it was transformed with BigM, we get the M values.
                for cons in transBlock.private_data().bigm_src:
                    m_values[cons] = self.get_M_value(cons)
        return m_values

    def get_largest_M_value(self, model):
        """Returns the largest M value for any constraint on the model.

        Parameters
        ----------
        model: A GDP model that has been transformed with BigM
        """
        return max(
            max(abs(m) for m in m_values if m is not None)
            for m_values in self.get_all_M_values_by_constraint(model).values()
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest

from pyomo.core.base.enums import SortComponents


class TestSortComponents(unittest.TestCase):
    def test_mappings(self):
        self.assertEqual(
            SortComponents(True),
            SortComponents.SORTED_INDICES | SortComponents.ALPHABETICAL,
        )
        self.assertEqual(SortComponents(False), SortComponents.UNSORTED)
        self.assertEqual(SortComponents(None), SortComponents.UNSORTED)
        with self.assertRaisesRegex(
            ValueError, r'(999 is not a valid SortComponents)|(invalid value 999)'
        ):
            SortComponents(999)
        # Note that enum maps False to 0 without hitting missing.  We
        # will explicitly test passing False to missing to cover a
        # "future proofing" logic branch
        self.assertEqual(SortComponents._missing_(False), SortComponents.UNSORTED)

    def test_sorter(self):
        self.assertEqual(SortComponents.sorter(), SortComponents.UNSORTED)
        self.assertEqual(
            SortComponents.sorter(True, False), SortComponents.ALPHABETICAL
        )
        self.assertEqual(
            SortComponents.sorter(False, True), SortComponents.SORTED_INDICES
        )
        self.assertEqual(
            SortComponents.sorter(True, True),
            SortComponents.ALPHABETICAL | SortComponents.SORTED_INDICES,
        )

    def test_methods(self):
        self.assertEqual(SortComponents.default(), SortComponents.UNSORTED)

        self.assertTrue(SortComponents.sort_names(SortComponents.sorter(True, True)))
        self.assertTrue(SortComponents.sort_names(SortComponents.sorter(True, False)))
        self.assertFalse(SortComponents.sort_names(SortComponents.sorter(False, True)))
        self.assertFalse(SortComponents.sort_names(SortComponents.sorter(False, False)))

        self.assertTrue(SortComponents.sort_indices(SortComponents.sorter(True, True)))
        self.assertFalse(
            SortComponents.sort_indices(SortComponents.sorter(True, False))
        )
        self.assertTrue(SortComponents.sort_indices(SortComponents.sorter(False, True)))
        self.assertFalse(
            SortComponents.sort_indices(SortComponents.sorter(False, False))
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Cut generation."""

from math import copysign
from pyomo.core import minimize, value
import pyomo.core.expr as EXPR
from pyomo.contrib.gdpopt.util import time_code
from pyomo.contrib.mcpp.pyomo_mcpp import McCormick as mc, MCPP_Error


def add_oa_cuts(
    target_model,
    dual_values,
    jacobians,
    objective_sense,
    mip_constraint_polynomial_degree,
    mip_iter,
    config,
    timing,
    cb_opt=None,
    linearize_active=True,
    linearize_violated=True,
):
    """Adds OA cuts.

    Generates and adds OA cuts (linearizes nonlinear constraints).
    For nonconvex problems, turn on 'config.add_slack'.
    Slack variables will always be used for nonlinear equality constraints.

    Parameters
    ----------
    target_model : Pyomo model
        The relaxed linear model.
    dual_values : list
        The value of the duals for each constraint.
    jacobians : ComponentMap
        Map nonlinear_constraint --> Map(variable --> jacobian of constraint w.r.t. variable).
    objective_sense : Int
        Objective sense of model.
    mip_constraint_polynomial_degree : Set
        The polynomial degrees of constraints that are regarded as linear.
    mip_iter : Int
        MIP iteration counter.
    config : ConfigBlock
        The specific configurations for MindtPy.
    cb_opt : SolverFactory, optional
        Gurobi_persistent solver, by default None.
    linearize_active : bool, optional
        Whether to linearize the active nonlinear constraints, by default True.
    linearize_violated : bool, optional
        Whether to linearize the violated nonlinear constraints, by default True.
    """
    with time_code(timing, 'OA cut generation'):
        for index, constr in enumerate(target_model.MindtPy_utils.constraint_list):
            # TODO: here the index is correlated to the duals, try if this can be fixed when temp duals are removed.
            if constr.body.polynomial_degree() in mip_constraint_polynomial_degree:
                continue

            constr_vars = list(EXPR.identify_variables(constr.body))
            jacs = jacobians

            # Equality constraint (makes the problem nonconvex)
            if (
                constr.has_ub()
                and constr.has_lb()
                and value(constr.lower) == value(constr.upper)
                and config.equality_relaxation
            ):
                sign_adjust = -1 if objective_sense == minimize else 1
                rhs = constr.lower
                if config.add_slack:
                    slack_var = target_model.MindtPy_utils.cuts.slack_vars.add()
                target_model.MindtPy_utils.cuts.oa_cuts.add(
                    expr=copysign(1, sign_adjust * dual_values[index])
                    * (
                        sum(
                            value(jacs[constr][var]) * (var - value(var))
                            for var in EXPR.identify_variables(constr.body)
                        )
                        + value(constr.body)
                        - rhs
                    )
                    - (slack_var if config.add_slack else 0)
                    <= 0
                )
                if (
                    config.single_tree
                    and config.mip_solver == 'gurobi_persistent'
                    and mip_iter > 0
                    and cb_opt is not None
                ):
                    cb_opt.cbLazy(
                        target_model.MindtPy_utils.cuts.oa_cuts[
                            len(target_model.MindtPy_utils.cuts.oa_cuts)
                        ]
                    )

            else:  # Inequality constraint (possibly two-sided)
                if (
                    constr.has_ub()
                    and (
                        linearize_active
                        and abs(constr.uslack()) < config.zero_tolerance
                    )
                    or (linearize_violated and constr.uslack() < 0)
                    or (config.linearize_inactive and constr.uslack() > 0)
                ) or (
                    'MindtPy_utils.objective_constr' in constr.name and constr.has_ub()
                ):
                    # always add the linearization for the epigraph of the objective
                    if config.add_slack:
                        slack_var = target_model.MindtPy_utils.cuts.slack_vars.add()

                    target_model.MindtPy_utils.cuts.oa_cuts.add(
                        expr=(
                            sum(
                                value(jacs[constr][var]) * (var - var.value)
                                for var in constr_vars
                            )
                            + value(constr.body)
                            - (slack_var if config.add_slack else 0)
                            <= value(constr.upper)
                        )
                    )
                    if (
                        config.single_tree
                        and config.mip_solver == 'gurobi_persistent'
                        and mip_iter > 0
                        and cb_opt is not None
                    ):
                        cb_opt.cbLazy(
                            target_model.MindtPy_utils.cuts.oa_cuts[
                                len(target_model.MindtPy_utils.cuts.oa_cuts)
                            ]
                        )

                if (
                    constr.has_lb()
                    and (
                        linearize_active
                        and abs(constr.lslack()) < config.zero_tolerance
                    )
                    or (linearize_violated and constr.lslack() < 0)
                    or (config.linearize_inactive and constr.lslack() > 0)
                ) or (
                    'MindtPy_utils.objective_constr' in constr.name and constr.has_lb()
                ):
                    if config.add_slack:
                        slack_var = target_model.MindtPy_utils.cuts.slack_vars.add()

                    target_model.MindtPy_utils.cuts.oa_cuts.add(
                        expr=(
                            sum(
                                value(jacs[constr][var]) * (var - var.value)
                                for var in constr_vars
                            )
                            + value(constr.body)
                            + (slack_var if config.add_slack else 0)
                            >= value(constr.lower)
                        )
                    )
                    if (
                        config.single_tree
                        and config.mip_solver == 'gurobi_persistent'
                        and mip_iter > 0
                        and cb_opt is not None
                    ):
                        cb_opt.cbLazy(
                            target_model.MindtPy_utils.cuts.oa_cuts[
                                len(target_model.MindtPy_utils.cuts.oa_cuts)
                            ]
                        )


def add_oa_cuts_for_grey_box(
    target_model, jacobians_model, config, objective_sense, mip_iter, cb_opt=None
):
    sign_adjust = -1 if objective_sense == minimize else 1
    if config.add_slack:
        slack_var = target_model.MindtPy_utils.cuts.slack_vars.add()
    for target_model_grey_box, jacobian_model_grey_box in zip(
        target_model.MindtPy_utils.grey_box_list,
        jacobians_model.MindtPy_utils.grey_box_list,
    ):
        jacobian_matrix = (
            jacobian_model_grey_box.get_external_model()
            .evaluate_jacobian_outputs()
            .toarray()
        )
        # Enumerate over values works well now. However, it might be stable if the values() method changes.
        for index, output in enumerate(target_model_grey_box.outputs.values()):
            dual_value = jacobians_model.dual[jacobian_model_grey_box][
                output.name.replace("outputs", "output_constraints")
            ]
            target_model.MindtPy_utils.cuts.oa_cuts.add(
                expr=copysign(1, sign_adjust * dual_value)
                * (
                    sum(
                        jacobian_matrix[index][var_index] * (var - value(var))
                        for var_index, var in enumerate(
                            target_model_grey_box.inputs.values()
                        )
                    )
                    - (output - value(output))
                )
                - (slack_var if config.add_slack else 0)
                <= 0
            )
    # TODO: gurobi_persistent currently does not support greybox model.
    # https://github.com/Pyomo/pyomo/issues/3000
    # if (
    #     config.single_tree
    #     and config.mip_solver == 'gurobi_persistent'
    #     and mip_iter > 0
    #     and cb_opt is not None
    # ):
    #     cb_opt.cbLazy(
    #         target_model.MindtPy_utils.cuts.oa_cuts[
    #             len(target_model.MindtPy_utils.cuts.oa_cuts)
    #         ]
    #     )


def add_ecp_cuts(
    target_model,
    jacobians,
    config,
    timing,
    linearize_active=True,
    linearize_violated=True,
):
    """Linearizes nonlinear constraints. Adds the cuts for the ECP method.

    Parameters
    ----------
    target_model : Pyomo model
        The relaxed linear model.
    jacobians : ComponentMap
        Map nonlinear_constraint --> Map(variable --> jacobian of constraint w.r.t. variable)
    config : ConfigBlock
        The specific configurations for MindtPy.
    timing : Timing
        Timing.
    linearize_active : bool, optional
        Whether to linearize the active nonlinear constraints, by default True.
    linearize_violated : bool, optional
        Whether to linearize the violated nonlinear constraints, by default True.
    """
    with time_code(timing, 'ECP cut generation'):
        for constr in target_model.MindtPy_utils.nonlinear_constraint_list:
            constr_vars = list(EXPR.identify_variables(constr.body))
            jacs = jacobians

            if constr.has_lb() and constr.has_ub():
                config.logger.warning(
                    'constraint {} has both a lower '
                    'and upper bound.'
                    '\n'.format(constr)
                )
                continue
            if constr.has_ub():
                try:
                    upper_slack = constr.uslack()
                except (ValueError, OverflowError) as e:
                    config.logger.error(e, exc_info=True)
                    config.logger.error(
                        'Constraint {} has caused either a '
                        'ValueError or OverflowError.'
                        '\n'.format(constr)
                    )
                    continue
                if (
                    (linearize_active and abs(upper_slack) < config.ecp_tolerance)
                    or (linearize_violated and upper_slack < 0)
                    or (config.linearize_inactive and upper_slack > 0)
                ):
                    if config.add_slack:
                        slack_var = target_model.MindtPy_utils.cuts.slack_vars.add()

                    target_model.MindtPy_utils.cuts.ecp_cuts.add(
                        expr=(
                            sum(
                                value(jacs[constr][var]) * (var - var.value)
                                for var in constr_vars
                            )
                            - (slack_var if config.add_slack else 0)
                            <= upper_slack
                        )
                    )

            if constr.has_lb():
                try:
                    lower_slack = constr.lslack()
                except (ValueError, OverflowError) as e:
                    config.logger.error(e, exc_info=True)
                    config.logger.error(
                        'Constraint {} has caused either a '
                        'ValueError or OverflowError.'
                        '\n'.format(constr)
                    )
                    continue
                if (
                    (linearize_active and abs(lower_slack) < config.ecp_tolerance)
                    or (linearize_violated and lower_slack < 0)
                    or (config.linearize_inactive and lower_slack > 0)
                ):
                    if config.add_slack:
                        slack_var = target_model.MindtPy_utils.cuts.slack_vars.add()

                    target_model.MindtPy_utils.cuts.ecp_cuts.add(
                        expr=(
                            sum(
                                value(jacs[constr][var]) * (var - var.value)
                                for var in constr_vars
                            )
                            + (slack_var if config.add_slack else 0)
                            >= -lower_slack
                        )
                    )


def add_no_good_cuts(target_model, var_values, config, timing, mip_iter=0, cb_opt=None):
    """Adds no-good cuts.

    This adds an no-good cuts to the no_good_cuts ConstraintList, which is not activated by default.
    However, it may be activated as needed in certain situations or for certain values of option flags.


    Parameters
    ----------
    target_model : Block
        The model to add no-good cuts to.
    var_values : list
        Variable values of the current solution, used to generate the cut.
    config : ConfigBlock
        The specific configurations for MindtPy.
    timing : Timing
        Timing.
    mip_iter : Int, optional
        MIP iteration counter.
    cb_opt : SolverFactory, optional
        Gurobi_persistent solver, by default None.

    Raises
    ------
    ValueError
        The value of binary variable is not 0 or 1.
    """
    if not config.add_no_good_cuts:
        return
    with time_code(timing, 'no_good cut generation'):
        config.logger.debug('Adding no-good cuts')

        m = target_model
        MindtPy = m.MindtPy_utils
        int_tol = config.integer_tolerance

        binary_vars = [v for v in MindtPy.variable_list if v.is_binary()]

        # copy variable values over
        for var, val in zip(MindtPy.variable_list, var_values):
            if not var.is_binary():
                continue
            var.set_value(val, skip_validation=True)

        # check to make sure that binary variables are all 0 or 1
        for v in binary_vars:
            if value(abs(v - 1)) > int_tol and value(abs(v)) > int_tol:
                raise ValueError(
                    'Binary {} = {} is not 0 or 1'.format(v.name, value(v))
                )

        if not binary_vars:  # if no binary variables, skip
            return

        int_cut = (
            sum(1 - v for v in binary_vars if value(abs(v - 1)) <= int_tol)
            + sum(v for v in binary_vars if value(abs(v)) <= int_tol)
            >= 1
        )

        MindtPy.cuts.no_good_cuts.add(expr=int_cut)
        if (
            config.single_tree
            and config.mip_solver == 'gurobi_persistent'
            and mip_iter > 0
            and cb_opt is not None
        ):
            cb_opt.cbLazy(
                target_model.MindtPy_utils.cuts.no_good_cuts[
                    len(target_model.MindtPy_utils.cuts.no_good_cuts)
                ]
            )


def add_affine_cuts(target_model, config, timing):
    """Adds affine cuts using MCPP.

    Parameters
    ----------
    config : ConfigBlock
        The specific configurations for MindtPy.
    timing : Timing
        Timing.
    """
    with time_code(timing, 'Affine cut generation'):
        m = target_model
        config.logger.debug('Adding affine cuts')
        counter = 0

        for constr in m.MindtPy_utils.nonlinear_constraint_list:
            vars_in_constr = list(EXPR.identify_variables(constr.body))
            if any(var.value is None for var in vars_in_constr):
                continue  # a variable has no values

            # mcpp stuff
            try:
                mc_eqn = mc(constr.body)
            except MCPP_Error as e:
                config.logger.error(e, exc_info=True)
                config.logger.error(
                    'Skipping constraint %s due to MCPP error' % (constr.name)
                )
                continue  # skip to the next constraint

            ccSlope = mc_eqn.subcc()
            cvSlope = mc_eqn.subcv()
            ccStart = mc_eqn.concave()
            cvStart = mc_eqn.convex()

            # check if the value of ccSlope and cvSlope is not Nan or inf. If so, we skip this.
            concave_cut_valid = True
            convex_cut_valid = True
            for var in vars_in_constr:
                if not var.fixed:
                    if ccSlope[var] == float('nan') or ccSlope[var] == float('inf'):
                        concave_cut_valid = False
                    if cvSlope[var] == float('nan') or cvSlope[var] == float('inf'):
                        convex_cut_valid = False
            # check if the value of ccSlope and cvSlope all equals zero. if so, we skip this.
            if not any(list(ccSlope.values())):
                concave_cut_valid = False
            if not any(list(cvSlope.values())):
                convex_cut_valid = False
            if ccStart == float('nan') or ccStart == float('inf'):
                concave_cut_valid = False
            if cvStart == float('nan') or cvStart == float('inf'):
                convex_cut_valid = False
            if not (concave_cut_valid or convex_cut_valid):
                continue

            ub_int = (
                min(value(constr.upper), mc_eqn.upper())
                if constr.has_ub()
                else mc_eqn.upper()
            )
            lb_int = (
                max(value(constr.lower), mc_eqn.lower())
                if constr.has_lb()
                else mc_eqn.lower()
            )

            aff_cuts = m.MindtPy_utils.cuts.aff_cuts
            if concave_cut_valid:
                concave_cut = (
                    sum(
                        ccSlope[var] * (var - var.value)
                        for var in vars_in_constr
                        if not var.fixed
                    )
                    + ccStart
                    >= lb_int
                )
                aff_cuts.add(expr=concave_cut)
                counter += 1
            if convex_cut_valid:
                convex_cut = (
                    sum(
                        cvSlope[var] * (var - var.value)
                        for var in vars_in_constr
                        if not var.fixed
                    )
                    + cvStart
                    <= ub_int
                )
                aff_cuts.add(expr=convex_cut)
                counter += 1

        config.logger.debug('Added %s affine cuts' % counter)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

from pyomo.common import DeveloperError
from pyomo.common.autoslots import AutoSlots
from pyomo.common.collections import ComponentMap
from pyomo.common.dependencies import numpy as np
from pyomo.common.dependencies.scipy import spatial
from pyomo.contrib.piecewise.piecewise_linear_expression import (
    PiecewiseLinearExpression,
)
from pyomo.contrib.piecewise.triangulations import (
    get_unordered_j1_triangulation,
    get_ordered_j1_triangulation,
    Triangulation,
)
from pyomo.core import Any, NonNegativeIntegers, value
from pyomo.core.base.block import BlockData, Block
from pyomo.core.base.component import ModelComponentFactory
from pyomo.core.base.expression import Expression
from pyomo.core.base.global_set import UnindexedComponent_index
from pyomo.core.base.indexed_component import UnindexedComponent_set
from pyomo.core.base.initializer import Initializer
import pyomo.core.expr as EXPR

# This is the default absolute tolerance in np.isclose... Not sure if it's
# enough, but we need to make sure that 'barely negative' values are assumed to
# be zero.
ZERO_TOLERANCE = 1e-8

logger = logging.getLogger(__name__)


class PiecewiseLinearFunctionData(BlockData):
    _Block_reserved_words = Any

    def __init__(self, component=None):
        BlockData.__init__(self, component)

        with self._declare_reserved_components():
            # map of PiecewiseLinearExpression objects to integer indices in
            # self._expressions
            self._expression_ids = ComponentMap()
            # index is monotonically increasing integer
            self._expressions = Expression(NonNegativeIntegers)
            self._transformed_exprs = ComponentMap()
            self._simplices = None
            # These will always be tuples, even when we only have one dimension.
            self._points = []
            self._linear_functions = []
            self._triangulation = None

    @property
    def triangulation(self):
        return self._triangulation

    def __call__(self, *args):
        """
        Returns a PiecewiseLinearExpression which is an instance of this
        function applied to the variables and/or constants specified in args.
        """
        if all(
            type(arg) in EXPR.native_types or not arg.is_potentially_variable()
            for arg in args
        ):
            # We need to actually evaluate
            return self._evaluate(*args)
        else:
            expr = PiecewiseLinearExpression(args, self)
            idx = len(self._expressions)
            self._expressions[idx] = expr
            self._expression_ids[expr] = idx
            return self._expressions[idx]

    def _evaluate(self, *args):
        # ESJ: This is a very inefficient implementation in high dimensions, but
        # for now we will just do a linear scan of the simplices.
        if self._simplices is None:
            raise RuntimeError(
                "Cannot evaluate PiecewiseLinearFunction--it "
                "appears it is not fully defined. (No simplices "
                "are stored.)"
            )

        pt = [value(arg) for arg in args]
        for simplex, func in zip(self._simplices, self._linear_functions):
            if self._pt_in_simplex(pt, simplex):
                return func(*args)

        raise ValueError(
            "Unsuccessful evaluation of PiecewiseLinearFunction "
            "'%s' at point (%s). Is the point in the function's "
            "domain?" % (self.name, ', '.join(str(arg) for arg in args))
        )

    def _pt_in_simplex(self, pt, simplex):
        dim = len(pt)
        if dim == 1:
            return (
                self._points[simplex[0]][0] <= pt[0]
                and self._points[simplex[1]][0] >= pt[0]
            )
        # Otherwise, we check if pt is a convex combination of the simplex's
        # extreme points
        A = np.ones((dim + 1, dim + 1))
        b = np.array([x for x in pt] + [1])
        for j, extreme_point in enumerate(simplex):
            for i, coord in enumerate(self._points[extreme_point]):
                A[i, j] = coord
        if np.linalg.det(A) == 0:
            # A is singular, so the system has no solutions
            return False
        else:
            lambdas = np.linalg.solve(A, b)
        for l in lambdas:
            if l < -ZERO_TOLERANCE:
                return False
        return True

    def _get_simplices_from_arg(self, simplices):
        self._simplices = []
        known_points = set()
        point_to_index = {}
        for simplex in simplices:
            extreme_pts = []
            for pt in simplex:
                if pt not in known_points:
                    known_points.add(pt)
                    if hasattr(pt, '__len__'):
                        self._points.append(pt)
                    else:
                        self._points.append((pt,))
                    point_to_index[pt] = len(self._points) - 1
                extreme_pts.append(point_to_index[pt])
            self._simplices.append(tuple(extreme_pts))

    def map_transformation_var(self, pw_expr, v):
        """
        Records on the PiecewiseLinearFunction object that the transformed
        form of the PiecewiseLinearExpression object pw_expr is the Var v.
        """
        if pw_expr not in self._expression_ids:
            raise DeveloperError(
                "ID of PiecewiseLinearExpression '%s' not in the _expression_ids "
                "dictionary of PiecewiseLinearFunction '%s'" % (pw_expr, self)
            )
        self._transformed_exprs[self._expressions[self._expression_ids[pw_expr]]] = v

    def get_transformation_var(self, pw_expr):
        """
        Returns the Var that replaced the PiecewiseLinearExpression 'pw_expr'
        after transformation, or None if 'pw_expr' has not been transformed.
        """
        if pw_expr in self._transformed_exprs:
            return self._transformed_exprs[pw_expr]
        else:
            return None


class _univariate_linear_functor(AutoSlots.Mixin):
    __slots__ = ('slope', 'intercept')

    def __init__(self, slope, intercept):
        self.slope = slope
        self.intercept = intercept

    def __call__(self, x):
        return self.slope * x + self.intercept


class _multivariate_linear_functor(AutoSlots.Mixin):
    __slots__ = ('normal',)

    def __init__(self, normal):
        self.normal = normal

    def __call__(self, *args):
        return sum(self.normal[i] * arg for i, arg in enumerate(args)) + self.normal[-1]


class _tabular_data_functor(AutoSlots.Mixin):
    __slots__ = ('tabular_data',)

    def __init__(self, tabular_data, tupleize=False):
        if not tupleize:
            self.tabular_data = tabular_data
        else:
            self.tabular_data = {(pt,): val for pt, val in tabular_data.items()}

    def __call__(self, *args):
        return self.tabular_data[args]


def _define_handler(handle_map, *key):
    def _wrapper(obj):
        assert key not in handle_map
        handle_map[key] = obj
        return obj

    return _wrapper


@ModelComponentFactory.register("Multidimensional piecewise linear function")
class PiecewiseLinearFunction(Block):
    """A piecewise linear function, which may be defined over an index.

    Can be specified in one of several ways:
        1) List of points and a nonlinear function to approximate. In
           this case, the points will be used to derive a triangulation
           of the part of the domain of interest, and a linear function
           approximating the given function will be calculated for each
           of the simplices in the triangulation. In this case, scipy is
           required (for multivariate functions).
        2) List of simplices and a nonlinear function to approximate. In
           this case, a linear function approximating the given function
           will be calculated for each simplex. For multivariate functions,
           numpy is required.
        3) List of simplices and list of functions that return linear function
           expressions. These are the desired piecewise functions
           corresponding to each simplex.
        4) Mapping of function values to points of the domain, allowing for
           the construction of a piecewise linear function from tabular data.

    Args:
        function: Nonlinear function to approximate: must be callable
        function_rule: Function that returns a nonlinear function to
            approximate for each index in an IndexedPiecewiseLinearFunction
        points: List of points in the same dimension as the domain of the
            function being approximated. Note that if the pieces of the
            function are specified this way, we require scipy.
        simplices: A list of lists of points, where each list specifies the
            extreme points of a a simplex over which the nonlinear function
            will be approximated as a linear function.
        linear_functions: A list of functions, each of which returns an
            expression for a linear function of the arguments.
        tabular_data: A dictionary mapping values of the nonlinear function
            to points in the domain
        triangulation (optional): An enum value of type Triangulation specifying
            how Pyomo should triangulate the function domain, or None. Behavior
            depends on how this piecewise-linear function is constructed:
            when constructed using methods (1) or (4) above, valid arguments
            are the members of Triangulation except Unknown or AssumeValid,
            and Pyomo will use that method to triangulate the domain and to tag
            the resulting PWLF. If no argument or None is passed, the default
            is Triangulation.Delaunay. When constructed using methods (2) or (3)
            above, valid arguments are only Triangulation.Unknown and
            Triangulation.AssumeValid. Pyomo will tag the constructed PWLF
            as specified, trusting the user in the case of AssumeValid.
            When no argument or None is passed, the default is
            Triangulation.Unknown
    """

    _ComponentDataClass = PiecewiseLinearFunctionData

    # Map 5-tuple of bool to handler: "(f, pts, simplices, linear_funcs,
    # tabular_data) : handler"
    _handlers = {}

    def __new__(cls, *args, **kwds):
        if cls is not PiecewiseLinearFunction:
            return super(PiecewiseLinearFunction, cls).__new__(cls)
        if not args or (args[0] is UnindexedComponent_set and len(args) == 1):
            return PiecewiseLinearFunction.__new__(ScalarPiecewiseLinearFunction)
        else:
            return IndexedPiecewiseLinearFunction.__new__(
                IndexedPiecewiseLinearFunction
            )

    def __init__(self, *args, **kwargs):
        _func_arg = kwargs.pop('function', None)
        _func_rule_arg = kwargs.pop('function_rule', None)
        _points_arg = kwargs.pop('points', None)
        _simplices_arg = kwargs.pop('simplices', None)
        _linear_functions = kwargs.pop('linear_functions', None)
        _tabular_data_arg = kwargs.pop('tabular_data', None)
        _tabular_data_rule_arg = kwargs.pop('tabular_data_rule', None)
        _triangulation_rule_arg = kwargs.pop('triangulation', None)

        kwargs.setdefault('ctype', PiecewiseLinearFunction)
        Block.__init__(self, *args, **kwargs)

        # This cannot be a rule.
        self._func = _func_arg
        self._func_rule = Initializer(_func_rule_arg)
        self._points_rule = Initializer(_points_arg, treat_sequences_as_mappings=False)
        self._simplices_rule = Initializer(
            _simplices_arg, treat_sequences_as_mappings=False
        )
        self._linear_funcs_rule = Initializer(
            _linear_functions, treat_sequences_as_mappings=False
        )
        self._tabular_data = _tabular_data_arg
        self._tabular_data_rule = Initializer(
            _tabular_data_rule_arg, treat_sequences_as_mappings=False
        )
        self._triangulation_rule = Initializer(
            _triangulation_rule_arg, treat_sequences_as_mappings=False
        )

    def _get_dimension_from_points(self, points):
        if len(points) < 1:
            raise ValueError(
                "Cannot construct PiecewiseLinearFunction from "
                "points list of length 0."
            )

        if hasattr(points[0], '__len__'):
            dimension = len(points[0])
        else:
            dimension = 1

        return dimension

    def _construct_simplices_from_multivariate_points(
        self, obj, parent, points, dimension
    ):
        if self._triangulation_rule is None:
            tri = Triangulation.Delaunay
        else:
            tri = self._triangulation_rule(parent, obj._index)
            if tri is None:
                tri = Triangulation.Delaunay

        if tri == Triangulation.Delaunay:
            try:
                triangulation = spatial.Delaunay(points)
            except (spatial.QhullError, ValueError) as error:
                logger.error("Unable to triangulate the set of input points.")
                raise
            obj._triangulation = tri
        elif tri == Triangulation.J1:
            triangulation = get_unordered_j1_triangulation(points, dimension)
            obj._triangulation = tri
        elif tri == Triangulation.OrderedJ1:
            triangulation = get_ordered_j1_triangulation(points, dimension)
            obj._triangulation = tri
        else:
            raise ValueError(
                "Invalid or unrecognized triangulation specified for '%s': %s"
                % (obj, tri)
            )

        # Get the points for the triangulation because they might not all be
        # there if any were coplanar.
        obj._points = [pt for pt in map(tuple, triangulation.points)]
        obj._simplices = []
        for simplex in triangulation.simplices:
            # For each simplex, check whether or not the simplex is
            # degenerate. If it is, we will just drop it.

            # We have n + 1 points in n dimensions.
            # We put them in a n x (n + 1) matrix: [p_0 p_1 ... p_n]
            points = triangulation.points[simplex].transpose()
            # The question is if they span R^n: We construct the square matrix
            # [p_1 - p_0  p_2 - p_1  ...  p_n - p_{n-1}] and check if it is full
            # rank. Note that we use numpy's matrix_rank function rather than
            # checking the determinant because matrix_rank will by default calculate a
            # tolerance based on the input to account for numerical errors in the
            # SVD computation.
            if tri in (Triangulation.J1, Triangulation.OrderedJ1):
                # Note: do not sort vertices from OrderedJ1, or it will break.
                # Non-ordered J1 is already sorted, though it doesn't matter.
                # Also, we don't need to check for degeneracy with simplices we
                # made ourselves.
                obj._simplices.append(tuple(simplex))
            elif (
                np.linalg.matrix_rank(
                    points[:, 1:]
                    - np.append(points[:, : dimension - 1], points[:, [0]], axis=1)
                )
                == dimension
            ):
                obj._simplices.append(tuple(sorted(simplex)))

        # It's possible that qhull dropped some points if there were numerical
        # issues with them (e.g., if they were redundant). We'll be polite and
        # tell the user:
        for pt in triangulation.coplanar:
            logger.info(
                "The Delaunay triangulation dropped the point with index "
                "%s from the triangulation." % pt[0]
            )

    # Call when constructing from simplices to allow use of AssumeValid and
    # ensure the user is not making mistakes
    def _check_and_set_triangulation_from_user(self, parent, obj):
        if self._triangulation_rule is None:
            tri = None
        else:
            tri = self._triangulation_rule(parent, obj._index)
        if tri is None or tri == Triangulation.Unknown:
            obj._triangulation = Triangulation.Unknown
        elif tri == Triangulation.AssumeValid:
            obj._triangulation = Triangulation.AssumeValid
        else:
            raise ValueError(
                f"Invalid or unrecognized triangulation tag specified for {obj} when"
                f" giving simplices: {tri}. Valid arguments when giving simplices are"
                " Triangulation.Unknown and Triangulation.AssumeValid."
            )

    def _construct_one_dimensional_simplices_from_points(self, obj, points):
        points.sort()
        obj._simplices = []
        for i in range(len(points) - 1):
            obj._simplices.append((i, i + 1))
            obj._points.append((points[i],))
        # Add the last one
        obj._points.append((points[-1],))

    @_define_handler(_handlers, True, True, False, False, False)
    def _construct_from_function_and_points(self, obj, parent, nonlinear_function):
        idx = obj._index

        points = self._points_rule(parent, idx)
        dimension = self._get_dimension_from_points(points)

        if dimension == 1:
            # This is univariate and we'll handle it separately in order to
            # avoid a dependence on scipy.
            self._construct_one_dimensional_simplices_from_points(obj, points)
            return self._construct_from_univariate_function_and_segments(
                obj, parent, nonlinear_function, segments_are_user_defined=False
            )

        self._construct_simplices_from_multivariate_points(
            obj, parent, points, dimension
        )
        return self._construct_from_function_and_simplices(
            obj, parent, nonlinear_function, simplices_are_user_defined=False
        )

    def _construct_from_univariate_function_and_segments(
        self, obj, parent, func, segments_are_user_defined=True
    ):
        # We can trust they are nicely ordered if we made them, otherwise anything goes.
        if segments_are_user_defined:
            self._check_and_set_triangulation_from_user(parent, obj)
        else:
            obj._triangulation = Triangulation.AssumeValid

        for idx1, idx2 in obj._simplices:
            x1 = obj._points[idx1][0]
            x2 = obj._points[idx2][0]
            y = {x: func(x) for x in [x1, x2]}
            slope = (y[x2] - y[x1]) / (x2 - x1)
            intercept = y[x1] - slope * x1
            obj._linear_functions.append(_univariate_linear_functor(slope, intercept))

        return obj

    @_define_handler(_handlers, True, False, True, False, False)
    def _construct_from_function_and_simplices(
        self, obj, parent, nonlinear_function, simplices_are_user_defined=True
    ):
        if obj._simplices is None:
            obj._get_simplices_from_arg(self._simplices_rule(parent, obj._index))
        simplices = obj._simplices

        if len(simplices) < 1:
            raise ValueError(
                "Cannot construct PiecewiseLinearFunction "
                "with empty list of simplices"
            )

        dimension = len(simplices[0]) - 1
        if dimension == 1:
            # Back to high school with us--this is univariate and we'll handle
            # it separately in order to avoid a kind of silly dependence on
            # numpy.
            return self._construct_from_univariate_function_and_segments(
                obj, parent, nonlinear_function, simplices_are_user_defined
            )

        # If we triangulated, then this tag was already set. If they provided it,
        # then check their arguments and set.
        if simplices_are_user_defined:
            self._check_and_set_triangulation_from_user(parent, obj)

        # evaluate the function at each of the points and form the homogeneous
        # system of equations
        A = np.ones((dimension + 2, dimension + 2))
        b = np.zeros(dimension + 2)
        b[-1] = 1

        for num_piece, simplex in enumerate(simplices):
            for i, pt_idx in enumerate(simplex):
                pt = obj._points[pt_idx]
                for j, val in enumerate(pt):
                    A[i, j] = val
                A[i, j + 1] = nonlinear_function(*pt)
            A[i + 1, :] = 0
            A[i + 1, dimension] = -1
            # This system has a solution unless there's a bug--we filtered the
            # simplices to make sure they are full-dimensional, so we know there
            # is a hyperplane that passes through these dimension + 1 points (and the
            # last equation scales it so that the coefficient for the output of
            # the nonlinear function dimension is -1, so we can just read off
            # the linear equation in the x space).
            try:
                normal = np.linalg.solve(A, b)
            except np.linalg.LinAlgError as e:
                logger.warning('LinAlgError: %s' % e)
                msg = (
                    "When calculating the hyperplane approximation over the simplex "
                    "with index %s, the matrix was unexpectedly singular. This "
                    "likely means that this simplex is degenerate" % num_piece
                )

                if simplices_are_user_defined:
                    raise ValueError(msg)
                # otherwise it's our fault, and I was hoping this is unreachable
                # code...
                raise DeveloperError(
                    msg
                    + " and that it should have been filtered out of the triangulation"
                )

            obj._linear_functions.append(_multivariate_linear_functor(normal))

        return obj

    @_define_handler(_handlers, False, False, True, True, False)
    def _construct_from_linear_functions_and_simplices(
        self, obj, parent, nonlinear_function
    ):
        # We know that we have simplices because else this handler wouldn't
        # have been called.
        obj._get_simplices_from_arg(self._simplices_rule(parent, obj._index))
        obj._linear_functions = [f for f in self._linear_funcs_rule(parent, obj._index)]
        self._check_and_set_triangulation_from_user(parent, obj)
        return obj

    @_define_handler(_handlers, False, False, False, False, True)
    def _construct_from_tabular_data(self, obj, parent, nonlinear_function):
        idx = obj._index

        tabular_data = self._tabular_data
        if tabular_data is None:
            tabular_data = self._tabular_data_rule(parent, idx)
        points = [pt for pt in tabular_data.keys()]
        dimension = self._get_dimension_from_points(points)

        if dimension == 1:
            # This is univariate and we'll handle it separately in order to
            # avoid a dependence on scipy.
            self._construct_one_dimensional_simplices_from_points(obj, points)
            return self._construct_from_univariate_function_and_segments(
                obj,
                parent,
                _tabular_data_functor(tabular_data, tupleize=True),
                segments_are_user_defined=False,
            )

        self._construct_simplices_from_multivariate_points(
            obj, parent, points, dimension
        )
        return self._construct_from_function_and_simplices(
            obj,
            parent,
            _tabular_data_functor(tabular_data),
            simplices_are_user_defined=False,
        )

    def _getitem_when_not_present(self, index):
        if index is None and not self.is_indexed():
            obj = self._data[index] = self
        else:
            obj = self._data[index] = self._ComponentDataClass(component=self)
        obj._index = index
        parent = obj.parent_block()

        # Get the nonlinear function, if we have one.
        nonlinear_function = None
        if self._func_rule is not None:
            nonlinear_function = self._func_rule(parent, index)
        elif self._func is not None:
            nonlinear_function = self._func

        handler = self._handlers.get(
            (
                nonlinear_function is not None,
                self._points_rule is not None,
                self._simplices_rule is not None,
                self._linear_funcs_rule is not None,
                self._tabular_data is not None or self._tabular_data_rule is not None,
            )
        )
        if handler is None:
            raise ValueError(
                "Unsupported set of arguments given for "
                "constructing PiecewiseLinearFunction. "
                "Expected a nonlinear function and a list"
                "of breakpoints, a nonlinear function and a list "
                "of simplices, a list of linear functions and "
                "a list of corresponding simplices, or a dictionary "
                "mapping points to nonlinear function values."
            )
        obj = handler(self, obj, parent, nonlinear_function)

        return obj


class ScalarPiecewiseLinearFunction(
    PiecewiseLinearFunctionData, PiecewiseLinearFunction
):
    def __init__(self, *args, **kwds):
        PiecewiseLinearFunctionData.__init__(self, self)
        PiecewiseLinearFunction.__init__(self, *args, **kwds)
        self._data[None] = self
        self._index = UnindexedComponent_index


class IndexedPiecewiseLinearFunction(PiecewiseLinearFunction):
    pass



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import json
import os
from os.path import join

import pyomo.common.unittest as unittest

from pyomo.common.fileutils import this_file_dir
from pyomo.common.log import LoggingIntercept
from pyomo.common.tempfiles import TempfileManager

from pyomo.opt import SolverFactory
from pyomo.core import ConcreteModel, Var, Objective, Constraint

currdir = this_file_dir()
deleteFiles = True

scip_available = False


class Test(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        global scip_available
        import pyomo.environ
        from pyomo.solvers.tests.solvers import test_solver_cases as _test_solver_cases

        scip_available = _test_solver_cases('scip', 'nl').available

    def setUp(self):
        if not scip_available:
            self.skipTest("The 'scipampl' command is not available")
        TempfileManager.push()

        self.scip = SolverFactory('scip', solver_io='nl')

        m = self.model = ConcreteModel()
        m.v = Var()
        m.o = Objective(expr=m.v)
        m.c = Constraint(expr=m.v >= 1)

    def tearDown(self):
        TempfileManager.pop(remove=deleteFiles or self.currentTestPassed())

    def compare_json(self, file1, file2):
        with open(file1, 'r') as out, open(file2, 'r') as txt:
            self.assertStructuredAlmostEqual(
                json.load(txt), json.load(out), abstol=1e-7, allow_second_superset=True
            )

    def test_version_scip(self):
        self.assertTrue(self.scip.version() is not None)
        self.assertTrue(type(self.scip.version()) is tuple)
        self.assertEqual(len(self.scip.version()), 4)

    def test_scip_solve_from_instance(self):
        # Test scip solve from a pyomo instance and load the solution
        results = self.scip.solve(self.model, suffixes=['.*'])
        # We don't want the test to care about which Scip version we are using
        self.model.solutions.store_to(results)
        results.Solution(0).Message = "Scip"
        results.Solver.Message = "Scip"
        results.Solver.Time = 0
        _out = TempfileManager.create_tempfile(".txt")
        results.write(filename=_out, times=False, format='json')
        self.compare_json(_out, join(currdir, "test_scip_solve_from_instance.baseline"))

    def test_scip_solve_from_instance_options(self):
        # Creating a dummy scip.set file in the cwd
        # will cover the code that prints a warning
        _cwd = os.getcwd()
        tmpdir = TempfileManager.create_tempdir()
        try:
            os.chdir(tmpdir)
            open(join(tmpdir, 'scip.set'), "w").close()
            # Test scip solve from a pyomo instance and load the solution
            with LoggingIntercept() as LOG:
                results = self.scip.solve(
                    self.model, suffixes=['.*'], options={"limits/softtime": 100}
                )
            self.assertRegex(
                LOG.getvalue().replace("\n", " "),
                r"A file named (.*) exists in the current working "
                r"directory, but SCIP options are being "
                r"set using a separate options file. The "
                r"options file \1 will be ignored.",
            )
        finally:
            os.chdir(_cwd)
        # We don't want the test to care about which Scip version we are using
        self.model.solutions.store_to(results)
        results.Solution(0).Message = "Scip"
        results.Solver.Message = "Scip"
        results.Solver.Time = 0
        _out = TempfileManager.create_tempfile(".txt")
        results.write(filename=_out, times=False, format='json')
        self.compare_json(_out, join(currdir, "test_scip_solve_from_instance.baseline"))

    def test_scip_solve_from_instance_with_reoptimization(self):
        # Test scip with re-optimization option enabled
        # This case changes the Scip output results which may break the results parser
        self.scip.options['reoptimization/enable'] = True
        self.test_scip_solve_from_instance()


if __name__ == "__main__":
    deleteFiles = False
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
from pyomo.environ import (
    SolverFactory,
    value,
    Var,
    Constraint,
    TransformationFactory,
    ConcreteModel,
    BooleanVar,
    LogicalConstraint,
    Block,
)
from pyomo.gdp import Disjunct, Disjunction
import pyomo.common.unittest as unittest
from pyomo.contrib.gdpopt.tests.four_stage_dynamic_model import build_model
from unittest.mock import MagicMock
from pyomo.core.expr.logical_expr import exactly
from pyomo.contrib.gdpopt.ldsda import GDP_LDSDA_Solver
from pyomo.opt import TerminationCondition as tc


class TestGDPoptLDSDA(unittest.TestCase):
    """Real unit tests for GDPopt"""

    @unittest.skipUnless(
        SolverFactory('gams').available(False)
        and SolverFactory('gams').license_is_valid(),
        "gams solver not available",
    )
    def test_solve_four_stage_dynamic_model(self):

        model = build_model(mode_transfer=True)

        # Discretize the model using dae.collocation
        discretizer = TransformationFactory('dae.collocation')
        discretizer.apply_to(model, nfe=10, ncp=3, scheme='LAGRANGE-RADAU')
        # We need to reconstruct the constraints in disjuncts after discretization.
        # This is a bug in Pyomo.dae. https://github.com/Pyomo/pyomo/issues/3101
        for disjunct in model.component_data_objects(ctype=Disjunct):
            for constraint in disjunct.component_objects(ctype=Constraint):
                constraint._constructed = False
                constraint.construct()

        for dxdt in model.component_data_objects(ctype=Var, descend_into=True):
            if 'dxdt' in dxdt.name:
                dxdt.setlb(-300)
                dxdt.setub(300)

        for direction_norm in ['L2', 'Linf']:
            result = SolverFactory('gdpopt.ldsda').solve(
                model,
                direction_norm=direction_norm,
                minlp_solver='gams',
                minlp_solver_args=dict(solver='ipopth'),
                starting_point=[1, 2],
                logical_constraint_list=[
                    model.mode_transfer_lc1,
                    model.mode_transfer_lc2,
                ],
                time_limit=100,
            )
            self.assertAlmostEqual(value(model.obj), -23.305325, places=4)


class TestLDSDALinearSearchUnit(unittest.TestCase):
    def test_line_search_tuple_unpacking(self):
        """
        Test that line_search correctly unpacks the (bool, float) tuple
        returned by _solve_GDP_subproblem.
        """
        # 1. Instantiate the solver class directly
        solver = GDP_LDSDA_Solver()

        # 2. Set up the fake internal state required for line_search
        solver.current_point = (0, 0)
        solver.best_direction = (1, 1)

        # 3. Mock the internal methods
        # check_valid_neighbor: Always say the neighbor is valid
        solver._check_valid_neighbor = MagicMock(return_value=True)

        # solve_GDP_subproblem: SIMULATE THE RETURN VALUES
        # Call 1: Returns (True, 10.0) -> Improvement found. Loop should continue.
        # Call 2: Returns (False, 10.0) -> No improvement. Loop SHOULD break.
        # If your fix works, the loop breaks here. If broken, it loops infinitely or crashes.
        solver._solve_GDP_subproblem = MagicMock(
            side_effect=[(True, 10.0), (False, 10.0)]
        )

        # 4. Run the method
        config = MagicMock()
        solver.line_search(config)

        # 5. Verify results
        # The solver should have moved exactly ONCE (from 0,0 to 1,1)
        self.assertEqual(solver.current_point, (1, 1))

        # The solver should have attempted to solve exactly TWICE
        # (Once for the success, once for the failure that stops the loop)
        self.assertEqual(solver._solve_GDP_subproblem.call_count, 2)


class TestLDSDAUnits(unittest.TestCase):
    """
    Unit tests for LDSDA behavior that exercise key control-flow,
    validation, and regression-sensitive code paths without relying
    on external solvers.
    """

    def setUp(self):
        self.solver = GDP_LDSDA_Solver()
        self.config = self.solver.CONFIG()

        self.model = ConcreteModel()
        self.model.util_block = Block()
        self.model.util_block.external_var_info_list = []
        self.model.util_block.parent_block = MagicMock(return_value=self.model)

    def test_any_termination_criterion_met(self):
        """
        Test the early termination of the main optimization loop.

        Verifies that the search loop breaks immediately if
        `any_termination_criterion_met` returns True, preventing further
        neighbor searches.
        """

        # 1. Mock critical methods to prevent real solving
        self.solver._solve_GDP_subproblem = MagicMock(return_value=(True, 0))
        # This forces the early-termination check in the main loop to trigger immediately,
        # causing the loop to break before any neighbor search is performed.
        self.solver.any_termination_criterion_met = MagicMock(return_value=True)
        self.solver.neighbor_search = MagicMock()
        # 2. Mock internal setup methods
        self.solver.get_external_information = MagicMock()
        self.solver._get_directions = MagicMock(return_value=[])

        # 3. FIX: Manually set the attribute that _get_external_information would have set
        self.solver.number_of_external_variables = 1
        self.config.starting_point = [0]

        # 4. Run the method
        self.solver._solve_gdp(self.model, self.config)

        # 5. Verify we hit the break (neighbor_search was skipped)
        self.solver.neighbor_search.assert_not_called()
        self.solver.any_termination_criterion_met.assert_called()

    def test_invalid_logical_constraint_type(self):
        """
        Test validation of logical constraint types.

        Verifies that a ValueError is raised if the logical constraints
        provided in the configuration are not instances of ExactlyExpression.
        """
        self.model.b = BooleanVar()
        self.model.lc = LogicalConstraint(expr=self.model.b.implies(True))

        self.config.logical_constraint_list = [self.model.lc]
        self.model.util_block.config_logical_constraint_list = [self.model.lc]

        with self.assertRaisesRegex(
            ValueError, "should be a list of ExactlyExpression"
        ):
            self.solver.get_external_information(self.model.util_block, self.config)

    def test_exactly_number_greater_than_one(self):
        """
        Test validation of the 'exactly' constraint argument.

        Verifies that a ValueError is raised if an ExactlyExpression is used
        with an integer argument greater than 1 (e.g., Exactly(2, ...)).
        """
        self.model.b1 = BooleanVar()
        self.model.b2 = BooleanVar()
        self.model.lc = LogicalConstraint(expr=exactly(2, self.model.b1, self.model.b2))

        self.config.logical_constraint_list = [self.model.lc]
        self.model.util_block.config_logical_constraint_list = [self.model.lc]

        with self.assertRaisesRegex(ValueError, "only works for exactly_number = 1"):
            self.solver.get_external_information(self.model.util_block, self.config)

    def test_starting_point_mismatch(self):
        """
        Test validation of the starting point dimensions.

        Verifies that a ValueError is raised if the length of the provided
        `starting_point` does not match the number of derived external variables.
        """
        self.model.b1 = BooleanVar()
        self.model.lc = LogicalConstraint(expr=exactly(1, self.model.b1))

        self.config.logical_constraint_list = [self.model.lc]
        self.model.util_block.config_logical_constraint_list = [self.model.lc]

        # Mismatch: 1 external var vs 2 starting points
        self.config.starting_point = [1, 2]

        with self.assertRaisesRegex(
            ValueError, "length of the provided starting point"
        ):
            self.solver.get_external_information(self.model.util_block, self.config)

    def test_disjunction_list_processing(self):
        """
        Test the processing of the disjunction list configuration.

        Verifies that providing a `disjunction_list` correctly populates the
        `external_var_info_list` with the appropriate bounds and metadata
        for the external variables.
        """
        self.model.d1 = Disjunct()
        self.model.d2 = Disjunct()
        self.model.disj = Disjunction(expr=[self.model.d1, self.model.d2])

        self.config.disjunction_list = [self.model.disj]
        self.model.util_block.config_disjunction_list = [self.model.disj]
        self.config.starting_point = [1]  # Correct length

        self.solver.get_external_information(self.model.util_block, self.config)

        # Verify it processed the disjunction
        self.assertEqual(len(self.model.util_block.external_var_info_list), 1)
        self.assertEqual(self.model.util_block.external_var_info_list[0].UB, 2)

    def test_neighbor_search_tiebreaker_logic(self):
        """
        Test the tie-breaking logic in the neighbor search.

        Verifies that when two neighbors offer improved objective values
        within the integer tolerance, the algorithm selects the neighbor
        that is Euclidean-farther from the current point.
        """
        self.solver.current_point = (0, 0)
        self.config.integer_tolerance = 1e-5

        # Manually define neighbors from (0, 0):
        # 1. (1, 0) - Distance 1
        # 2. (1, 1) - Distance sqrt(2) (further away)
        self.solver.directions = [(1, 0), (1, 1)]
        self.solver._check_valid_neighbor = MagicMock(return_value=True)

        # Mock subproblems to return IDENTICAL objectives
        # This forces the code to check the distance to break the tie
        self.solver._solve_GDP_subproblem = MagicMock(
            side_effect=[(True, 100.0), (True, 100.0)]
        )

        self.solver.neighbor_search(self.config)

        # It should pick (1,1) because it is further away (Tiebreaker rule)
        self.assertEqual(self.solver.current_point, (1, 1))

    def test_handle_subproblem_result_none(self):
        """
        Test handling of None results from subproblems.

        Verifies that `_handle_subproblem_result` returns False (indicating
        no improvement) if the subproblem solver returns None.
        """
        result = self.solver._handle_subproblem_result(
            None, (False, None), None, None, None
        )
        self.assertFalse(result)

    def test_handle_subproblem_result_termination_failure(self):
        """
        Test handling of failed termination conditions.

        Verifies that `_handle_subproblem_result` returns False if the solver
        termination condition is not in the set of successful statuses
        (e.g., optimal, feasible).
        """
        # 1. Create a mock result object
        mock_result = MagicMock()

        # 2. Set termination condition to something NOT in the 'success' set
        # The code checks for: optimal, feasible, globallyOptimal, locallyOptimal,
        # maxTimeLimit, maxIterations, maxEvaluations.
        # We use 'infeasible' or 'error' to trigger the False return.
        mock_result.solver.termination_condition = tc.error

        # 3. Run the method
        # We don't need real model objects for the other args since the code exits early
        result = self.solver._handle_subproblem_result(
            mock_result, None, None, None, None
        )

        # 4. Verify it returns False (because the solver failed)
        self.assertFalse(result)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#

import re
import os
from os.path import abspath, dirname, join

currdir = dirname(abspath(__file__))

from filecmp import cmp
import pyomo.common.unittest as unittest

from pyomo.opt import check_available_solvers
from pyomo.scripting.pyomo_main import main
from pyomo.core import (
    AbstractModel,
    ConcreteModel,
    Block,
    Set,
    Param,
    Var,
    Objective,
    Constraint,
    Reals,
    display,
)
from pyomo.common.tee import capture_output

from io import StringIO


def rule1(model):
    return (1, model.x + model.y[1], 2)


def rule2(model, i):
    return (1, model.x + model.y[1] + i, 2)


solvers = None


class PyomoModel(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        global solvers
        import pyomo.environ

        solvers = check_available_solvers('glpk', 'cplex')

    def test_construct(self):
        model = AbstractModel()
        model.a = Set(initialize=[1, 2, 3])
        model.A = Param(initialize=1)
        model.B = Param(model.a)
        model.x = Var(initialize=1, within=Reals, dense=False)
        model.y = Var(model.a, initialize=1, within=Reals, dense=False)
        model.obj = Objective(rule=lambda model: model.x + model.y[1])
        model.obj2 = Objective(model.a, rule=lambda model, i: i + model.x + model.y[1])
        model.con = Constraint(rule=rule1)
        model.con2 = Constraint(model.a, rule=rule2)
        instance = model.create_instance()
        expr = instance.x + 1

        OUTPUT = open(join(currdir, "display.out"), "w")
        display(instance, ostream=OUTPUT)
        display(instance.obj, ostream=OUTPUT)
        display(instance.x, ostream=OUTPUT)
        display(instance.con, ostream=OUTPUT)
        OUTPUT.write(expr.to_string())
        model = AbstractModel()
        instance = model.create_instance()
        display(instance, ostream=OUTPUT)
        OUTPUT.close()
        try:
            display(None)
            self.fail("test_construct - expected TypeError")
        except TypeError:
            pass
        _out, _txt = join(currdir, "display.out"), join(currdir, "display.txt")
        self.assertTrue(cmp(_out, _txt), msg="Files %s and %s differ" % (_out, _txt))

    def test_construct2(self):
        model = AbstractModel()
        model.a = Set(initialize=[1, 2, 3])
        model.A = Param(initialize=1)
        model.B = Param(model.a)
        model.x = Var(initialize=1, within=Reals, dense=True)
        model.y = Var(model.a, initialize=1, within=Reals, dense=True)
        model.obj = Objective(rule=lambda model: model.x + model.y[1])
        model.obj2 = Objective(model.a, rule=lambda model, i: i + model.x + model.y[1])
        model.con = Constraint(rule=rule1)
        model.con2 = Constraint(model.a, rule=rule2)
        instance = model.create_instance()
        expr = instance.x + 1

        OUTPUT = open(join(currdir, "display2.out"), "w")
        display(instance, ostream=OUTPUT)
        display(instance.obj, ostream=OUTPUT)
        display(instance.x, ostream=OUTPUT)
        display(instance.con, ostream=OUTPUT)
        OUTPUT.write(expr.to_string())
        model = AbstractModel()
        instance = model.create_instance()
        display(instance, ostream=OUTPUT)
        OUTPUT.close()
        try:
            display(None)
            self.fail("test_construct - expected TypeError")
        except TypeError:
            pass
        _out, _txt = join(currdir, "display2.out"), join(currdir, "display2.txt")
        self.assertTrue(cmp(_out, _txt), msg="Files %s and %s differ" % (_out, _txt))


class PyomoBadModels(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        global solvers
        import pyomo.environ

        solvers = check_available_solvers('glpk', 'cplex')

    def pyomo(self, cmd, **kwargs):
        args = ['solve'] + re.split('[ ]+', cmd)
        out = kwargs.get('file', None)
        if out is None:
            out = StringIO()
        with capture_output(out):
            os.chdir(currdir)
            output = main(args)
        if not 'file' in kwargs:
            return output.getvalue()
        return output

    def test_uninstantiated_model_linear(self):
        """Run pyomo with "bad" model file.  Should fail gracefully, with
        a perhaps useful-to-the-user message."""
        if not 'glpk' in solvers:
            self.skipTest("glpk solver is not available")
        return  # ignore for now
        base = '%s/test_uninstantiated_model' % currdir
        fout, fbase = join(base, '_linear.out'), join(base, '.txt')
        self.pyomo('uninstantiated_model_linear.py', file=fout)
        self.assertTrue(cmp(fout, fbase), msg="Files %s and %s differ" % (fout, fbase))

    def test_uninstantiated_model_quadratic(self):
        """Run pyomo with "bad" model file.  Should fail gracefully, with
        a perhaps useful-to-the-user message."""
        if not 'cplex' in solvers:
            self.skipTest("The 'cplex' executable is not available")
        return  # ignore for now
        base = '%s/test_uninstantiated_model' % currdir
        fout, fbase = join(base, '_quadratic.out'), join(base, '.txt')
        self.pyomo('uninstantiated_model_quadratic.py --solver=cplex', file=fout)
        self.assertTrue(cmp(fout, fbase), msg="Files %s and %s differ" % (fout, fbase))


class TestApplyIndexedRule(unittest.TestCase):
    def test_rules_with_None_in_set(self):
        def noarg_rule(b):
            b.args = ()

        def onearg_rule(b, i):
            b.args = (i,)

        def twoarg_rule(b, i, j):
            b.args = (i, j)

        m = ConcreteModel()
        m.b1 = Block(rule=noarg_rule)
        self.assertEqual(m.b1.args, ())

        m.b2 = Block([None], rule=onearg_rule)
        self.assertEqual(m.b2[None].args, (None,))

        m.b3 = Block([(None, 1)], rule=twoarg_rule)
        self.assertEqual(m.b3[None, 1].args, ((None, 1)))


class TestComponent(unittest.TestCase):
    def test_getname(self):
        m = ConcreteModel()
        m.b = Block()
        m.b.v = Var()
        self.assertEqual(m.b.v.getname(fully_qualified=True, relative_to=m.b), 'v')

    def test_getname_error(self):
        m = ConcreteModel()
        m.b = Block()
        m.b.v = Var()
        m.c = Block()
        self.assertRaises(
            RuntimeError, m.b.v.getname, fully_qualified=True, relative_to=m.c
        )


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for Elements of a Model
#
# TestSimpleCon                Class for testing single constraint
# TestArrayCon                Class for testing array of constraint
#

import os
from os.path import abspath, dirname

currdir = dirname(abspath(__file__)) + os.sep

import pyomo.common.unittest as unittest

from pyomo.environ import AbstractModel, Param, Var, Constraint, value


class TestMutable(unittest.TestCase):
    def test_mutable_constraint_upper(self):
        model = AbstractModel()
        model.P = Param(initialize=2.0, mutable=True)
        model.X = Var()

        def constraint_rule(m):
            return m.X <= m.P

        model.C = Constraint(rule=constraint_rule)

        instance = model.create_instance()

        self.assertEqual(value(instance.C.upper), 2.0)

        instance.P = 4.0

        self.assertEqual(value(instance.C.upper), 4.0)

    def test_mutable_constraint_lower(self):
        model = AbstractModel()
        model.Q = Param(initialize=2.0, mutable=True)
        model.X = Var()

        def constraint_rule(m):
            return m.X >= m.Q

        model.C = Constraint(rule=constraint_rule)

        instance = model.create_instance()

        self.assertEqual(value(instance.C.lower), 2.0)

        instance.Q = 4.0

        self.assertEqual(value(instance.C.lower), 4.0)

    def test_mutable_constraint_both(self):
        model = AbstractModel()
        model.P = Param(initialize=4.0, mutable=True)
        model.Q = Param(initialize=2.0, mutable=True)
        model.X = Var()

        def constraint_rule(m):
            return (m.Q, m.X, m.P)

        model.C = Constraint(rule=constraint_rule)

        instance = model.create_instance()

        self.assertEqual(value(instance.C.lower), 2.0)
        self.assertEqual(value(instance.C.upper), 4.0)

        instance.P = 8.0
        instance.Q = 1.0

        self.assertEqual(value(instance.C.lower), 1.0)
        self.assertEqual(value(instance.C.upper), 8.0)

    def test_mutable_var_bounds_lower(self):
        model = AbstractModel()
        model.P = Param(initialize=2.0, mutable=True)
        model.X = Var(bounds=(model.P, None))

        instance = model.create_instance()

        self.assertEqual(instance.X.bounds, (2.0, None))

        instance.P = 4.0

        self.assertEqual(instance.X.bounds, (4.0, None))

    def test_mutable_var_bounds_upper(self):
        model = AbstractModel()
        model.Q = Param(initialize=2.0, mutable=True)
        model.X = Var(bounds=(model.Q, None))

        instance = model.create_instance()

        self.assertEqual(instance.X.bounds, (2.0, None))

        instance.Q = 4.0

        self.assertEqual(instance.X.bounds, (4.0, None))

    def test_mutable_var_bounds_both(self):
        model = AbstractModel()
        model.P = Param(initialize=4.0, mutable=True)
        model.Q = Param(initialize=2.0, mutable=True)
        model.X = Var(bounds=(model.P, model.Q))

        instance = model.create_instance()

        self.assertEqual(value(instance.X.lb), 4.0)
        self.assertEqual(value(instance.X.ub), 2.0)

        instance.P = 8.0
        instance.Q = 1.0

        self.assertEqual(value(instance.X.lb), 8.0)
        self.assertEqual(value(instance.X.ub), 1.0)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for Discrete Variable Transformations

import pyomo.common.unittest as unittest

from pyomo.core.base import VarCollector
from pyomo.environ import (
    ConcreteModel,
    Var,
    Constraint,
    Objective,
    Suffix,
    Binary,
    TransformationFactory,
    SolverFactory,
    Reals,
    Block,
    Integers,
    value,
)
from pyomo.opt import check_available_solvers

solvers = check_available_solvers('cplex', 'gurobi', 'glpk')


def _generateModel():
    model = ConcreteModel()
    model.x = Var(within=Binary)
    model.y = Var()
    model.c1 = Constraint(expr=model.y >= model.x)
    model.c2 = Constraint(expr=model.y >= 1.5 - model.x)
    model.obj = Objective(expr=model.y)
    model.dual = Suffix(direction=Suffix.IMPORT_EXPORT)
    return model


def _make_hierarchical_model():
    m = ConcreteModel()
    m.y = Var(domain=Binary)
    m.b = Block()
    m.b.x = Var([1, 2], bounds=(2, 45), domain=Integers)
    m.b.y = Var(domain=Binary)
    m.b.c = Constraint(expr=m.b.x[1] * m.y <= 23)

    return m


class Test(unittest.TestCase):
    @unittest.skipIf(len(solvers) == 0, "LP/MIP solver not available")
    def test_solve_relax_transform(self):
        s = SolverFactory(solvers[0])
        m = _generateModel()
        self.assertIs(m.x.domain, Binary)
        self.assertEqual(m.x.lb, 0)
        self.assertEqual(m.x.ub, 1)
        s.solve(m)
        self.assertEqual(len(m.dual), 0)

        TransformationFactory('core.relax_integer_vars').apply_to(m)
        self.assertIs(m.x.domain, Reals)
        self.assertEqual(m.x.lb, 0)
        self.assertEqual(m.x.ub, 1)
        s.solve(m)
        self.assertEqual(len(m.dual), 2)
        self.assertAlmostEqual(m.dual[m.c1], -0.5, 4)
        self.assertAlmostEqual(m.dual[m.c2], -0.5, 4)

    def test_reverse_relax_integer_vars(self):
        m = _generateModel()
        lp_relax = TransformationFactory('core.relax_integer_vars')
        reverse = lp_relax.apply_to(m)
        self.assertIs(m.x.domain, Reals)
        self.assertIs(m.y.domain, Reals)
        self.assertEqual(m.x.lb, 0)
        self.assertEqual(m.x.ub, 1)
        self.assertIsNone(m.y.lb)
        self.assertIsNone(m.y.ub)

        lp_relax.apply_to(m, reverse=reverse)
        self.assertIs(m.x.domain, Binary)
        self.assertIs(m.y.domain, Reals)
        self.assertEqual(m.x.lb, 0)
        self.assertEqual(m.x.ub, 1)
        self.assertIsNone(m.y.lb)
        self.assertIsNone(m.y.ub)

    def test_relax_integer_vars_block_targets(self):
        m = _make_hierarchical_model()
        TransformationFactory('core.relax_integer_vars').apply_to(m, targets=m.b)
        for i in [1, 2]:
            self.assertIs(m.b.x[i].domain, Reals)
            self.assertEqual(m.b.x[i].lb, 2)
            self.assertEqual(m.b.x[i].ub, 45)
        self.assertIs(m.b.y.domain, Reals)
        self.assertEqual(m.b.y.lb, 0)
        self.assertEqual(m.b.y.ub, 1)

        self.assertIs(m.y.domain, Binary)
        self.assertEqual(m.y.lb, 0)
        self.assertEqual(m.y.ub, 1)

    def test_relax_integer_vars_var_data_targets(self):
        m = _make_hierarchical_model()
        TransformationFactory('core.relax_integer_vars').apply_to(
            m, targets=[m.b.x[1], m.y]
        )
        # transformed
        self.assertIs(m.b.x[1].domain, Reals)
        self.assertEqual(m.b.x[1].lb, 2)
        self.assertEqual(m.b.x[1].ub, 45)
        # not transformed
        self.assertIs(m.b.x[2].domain, Integers)
        self.assertEqual(m.b.x[2].lb, 2)
        self.assertEqual(m.b.x[2].ub, 45)
        # not transformed
        self.assertIs(m.b.y.domain, Binary)
        self.assertEqual(m.b.y.lb, 0)
        self.assertEqual(m.b.y.ub, 1)
        # transformed
        self.assertIs(m.y.domain, Reals)
        self.assertEqual(m.y.lb, 0)
        self.assertEqual(m.y.ub, 1)

    def test_relax_integer_vars_indexed_var_targets(self):
        m = _make_hierarchical_model()
        TransformationFactory('core.relax_integer_vars').apply_to(m, targets=m.b.x)
        # transformed
        for i in [1, 2]:
            self.assertIs(m.b.x[i].domain, Reals)
            self.assertEqual(m.b.x[i].lb, 2)
            self.assertEqual(m.b.x[i].ub, 45)
        # not transformed
        self.assertIs(m.b.y.domain, Binary)
        self.assertEqual(m.b.y.lb, 0)
        self.assertEqual(m.b.y.ub, 1)
        # not transformed
        self.assertIs(m.y.domain, Binary)
        self.assertEqual(m.y.lb, 0)
        self.assertEqual(m.y.ub, 1)

    def test_relax_integer_vars_vars_from_expressions(self):
        m = _make_hierarchical_model()
        TransformationFactory('core.relax_integer_vars').apply_to(
            m.b, var_collector=VarCollector.FromExpressions
        )
        # transformed
        self.assertIs(m.b.x[1].domain, Reals)
        self.assertEqual(m.b.x[1].lb, 2)
        self.assertEqual(m.b.x[1].ub, 45)
        # not transformed
        self.assertIs(m.b.x[2].domain, Integers)
        self.assertEqual(m.b.x[2].lb, 2)
        self.assertEqual(m.b.x[2].ub, 45)
        # not transformed
        self.assertIs(m.b.y.domain, Binary)
        self.assertEqual(m.b.y.lb, 0)
        self.assertEqual(m.b.y.ub, 1)
        # transformed
        self.assertIs(m.y.domain, Reals)
        self.assertEqual(m.y.lb, 0)
        self.assertEqual(m.y.ub, 1)

    def test_relax_integer_vars_ignore_deactivated_blocks(self):
        m = _make_hierarchical_model()
        m.b.deactivate()
        TransformationFactory('core.relax_integer_vars').apply_to(
            m, transform_deactivated_blocks=False
        )
        # not transformed
        for i in [1, 2]:
            self.assertIs(m.b.x[i].domain, Integers)
            self.assertEqual(m.b.x[i].lb, 2)
            self.assertEqual(m.b.x[i].ub, 45)
        # not transformed
        self.assertIs(m.b.y.domain, Binary)
        self.assertEqual(m.b.y.lb, 0)
        self.assertEqual(m.b.y.ub, 1)
        # transformed
        self.assertIs(m.y.domain, Reals)
        self.assertEqual(m.y.lb, 0)
        self.assertEqual(m.y.ub, 1)

        m = _make_hierarchical_model()
        m.obj = Objective(expr=m.b.x[2])
        m.b.deactivate()
        TransformationFactory('core.relax_integer_vars').apply_to(
            m,
            transform_deactivated_blocks=False,
            var_collector=VarCollector.FromExpressions,
        )
        # not transformed
        self.assertIs(m.b.x[1].domain, Integers)
        self.assertEqual(m.b.x[1].lb, 2)
        self.assertEqual(m.b.x[1].ub, 45)
        # transformed
        self.assertIs(m.b.x[2].domain, Reals)
        self.assertEqual(m.b.x[2].lb, 2)
        self.assertEqual(m.b.x[2].ub, 45)
        # not transformed
        self.assertIs(m.b.y.domain, Binary)
        self.assertEqual(m.b.y.lb, 0)
        self.assertEqual(m.b.y.ub, 1)
        # not transformed
        self.assertIs(m.y.domain, Binary)
        self.assertEqual(m.y.lb, 0)
        self.assertEqual(m.y.ub, 1)

    def test_relax_integer_vars_fixed_vars(self):
        m = _make_hierarchical_model()
        m.y.fix(0)
        m.b.y.fix(1)
        reverse = TransformationFactory('core.relax_integer_vars').apply_to(m)

        # change the domain, but don't unfix
        self.assertIs(m.y.domain, Reals)
        self.assertEqual(m.y.lb, 0)
        self.assertEqual(m.y.ub, 1)
        self.assertEqual(value(m.y), 0)
        self.assertTrue(m.y.fixed)

        self.assertIs(m.b.y.domain, Reals)
        self.assertEqual(m.b.y.lb, 0)
        self.assertEqual(m.b.y.ub, 1)
        self.assertEqual(value(m.b.y), 1)
        self.assertTrue(m.b.y.fixed)

        # transformed
        for i in [1, 2]:
            self.assertIs(m.b.x[i].domain, Reals)
            self.assertEqual(m.b.x[i].lb, 2)
            self.assertEqual(m.b.x[i].ub, 45)

        # reverse and make sure fixed guys are still fixed
        TransformationFactory('core.relax_integer_vars').apply_to(m, reverse=reverse)
        self.assertIs(m.y.domain, Binary)
        self.assertEqual(m.y.lb, 0)
        self.assertEqual(m.y.ub, 1)
        self.assertEqual(value(m.y), 0)
        self.assertTrue(m.y.fixed)

        self.assertIs(m.b.y.domain, Binary)
        self.assertEqual(m.b.y.lb, 0)
        self.assertEqual(m.b.y.ub, 1)
        self.assertEqual(value(m.b.y), 1)
        self.assertTrue(m.b.y.fixed)

    @unittest.skipIf(len(solvers) == 0, "LP/MIP solver not available")
    def test_solve_fix_transform(self):
        s = SolverFactory(solvers[0])
        m = _generateModel()
        self.assertIs(m.x.domain, Binary)
        self.assertEqual(m.x.lb, 0)
        self.assertEqual(m.x.ub, 1)
        s.solve(m)
        m.pprint()
        self.assertEqual(len(m.dual), 0)

        TransformationFactory('core.fix_discrete').apply_to(m)
        self.assertIs(m.x.domain, Binary)
        self.assertEqual(m.x.lb, 0)
        self.assertEqual(m.x.ub, 1)
        s.solve(m)
        self.assertEqual(len(m.dual), 2)
        self.assertAlmostEqual(m.dual[m.c1], -1, 4)
        self.assertAlmostEqual(m.dual[m.c2], 0, 4)


if __name__ == "__main__":
    unittest.main()
